"""Module for calculating values for a tornadoplot"""

import pandas as pd


def real_mask(dfr, start, end):
    """Creates mask for which realisations to calc from"""
    mask = (dfr.REAL >= start) & (dfr.REAL <= end)
    return mask


def check_selector(resultfile, selector):
    """Checks whether selector is in resultfile headers"""
    if selector not in resultfile.columns:
        raise ValueError("Did not find ", selector, " as column in", "resultfile")


def check_selection(resultfile, selector, selection):
    """Checks whether selction is in resultfile values
    in the column selector"""
    for sel in selection:
        if sel not in resultfile[selector].values:
            if str(sel).strip("[]").lower() == "total":
                raise ValueError("Use 'all' to sum all", selector, "not 'total'")
            raise ValueError(
                "Selection ", sel, " not found in column", selector, "in resultfile"
            )
        resultfile[selector].values.astype(str)


def check_response(resultfile, response):
    """Checks whether respones in in resultfile"""
    if response not in resultfile.columns:
        raise ValueError("Did not find ", response, " as column in", "resultfile")


def cut_by_ref(tornadotable, refname):
    """Removes sensitivities smaller than reference sensitivity from table"""
    maskref = tornadotable.sensname == refname
    reflow = tornadotable[maskref].low.abs()
    refhigh = tornadotable[maskref].high.abs()
    refmax = max(float(reflow), float(refhigh))
    dfr_filtered = tornadotable.loc[
        (tornadotable["sensname"] == refname)
        | (
            (tornadotable["low"].abs() >= refmax)
            | (tornadotable["high"].abs() >= refmax)
        )
    ]
    return dfr_filtered


def sort_by_max(tornadotable):
    """Sorts table based on max(abs('low', 'high'))"""
    tornadotable["max"] = (
        tornadotable[["low", "high"]]
        .apply(lambda x: max(x.min(), x.max(), key=abs), axis=1)
        .abs()
    )
    df_sorted = tornadotable.sort_values("max", ascending=True)
    df_sorted.drop(["max"], axis=1, inplace=True)
    return df_sorted


def calc_tornadoinput(
    designsummary,
    resultfile,
    response,
    selectors,
    selection,
    reference="rms_seed",
    scale="percentage",
    cutbyref=False,
    sortsens=True,
):
    """
     Calculates input values for a tornadoplot for one response
     and one design set up

    Args:
        designsummary (pd.DataFrame): Summary of designmatrix as output
            from summarize_design.
        resultfile (pd.DataFrame): DataFrame with collected results
            for different realizations where
            'REAL' is one of the columns.
        response (str): name of response in resultfile to plot tornado for
        selectors (list): List of strings. Selectors to choose/filter on
        selections (list): List of lists. Values to filter on for each selector.
        reference (str): Specifying what the reference is for
            the tornado plots. Defaults to 'rms_seed'.
            Valid choices are a sensname or a
            realisation number.
        scale (str): Whether to plot absolute numbers or percentage
            compared to reference mean.  Defaults to 'percentage'.
            Valid choices are 'percentage' and 'absolute'.
        cutbyref (bool): If True sensitivities smaller than ref
            are excluded. Defaults to False.
        sortsens (bool): If True sensitivities are sorted so that
            largest comes first. Defaults to True.

    Returns:
        tuple:

            * (pd.DataFrame): tornadoinput for DataFrame for
                                  a tornado plot
            * (int): Average response value for realizations in
                     reference sensitivity.

    Example::
        >>> import pandas as pd
        >>> from fmu.tools.sensitivities import calc_tornadoplot
        >>> designtable = pd.read_csv('designsummary.csv')
        >>> results = pd.read_csv('resultfile.csv')
        >>> response = 'STOIIP_OIL'
        >>> selectors = ['ZONE', 'REGION']
        >>> selection = [['Nansen','Larsson'], ['SegmentA']]
        >>> reference = 'rms_seed'
        >>> scale = 'percentage'
        >>> cutbyref = True
        >>> sortsens = False
        >>> (tornadotable, ref_value) = calc_tornadoinput(
            designtable, results, response, selectors,
            selection, reference, scale, cutbyref, sortsens)

    """
    # Check that chosen response exists in resultfile
    check_response(resultfile, response)

    # Filter on chosen selectors(column names) and selections(values)
    for itr, sel in enumerate(selection):
        header = selectors[itr]
        sel_stripped = str(sel).strip("[]").lower()
        check_selector(resultfile, selectors[itr])
        if sel_stripped != "'all'":
            resultfile[header] = resultfile[header].astype(str)
            check_selection(resultfile, selectors[itr], sel)
            resultfile = resultfile[(resultfile[header].isin(sel))]

    # summing over chosen selections for each realisation
    headers = ["REAL"] + selectors
    resultcopy = resultfile.copy()
    resultcopy.set_index(headers, inplace=True)
    resultcopy.sort_index(inplace=True)
    dfr_summed = resultcopy.groupby(["REAL"]).sum(min_count=1)
    dfr_summed.reset_index(inplace=True)

    # Creates empty dfr to store stats in, webportal tornadoplot input format
    tornadoinput = pd.DataFrame(
        columns=[
            "sensno",
            "sensname",
            "low",
            "high",
            "leftlabel",
            "rightlabel",
            "numreal1",
            "numreal2",
        ]
    )

    # Calculate mean for reference
    if reference in designsummary["sensname"].values:
        startreal = int(
            designsummary[designsummary["sensname"] == reference]["startreal1"]
        )
        endreal = int(designsummary[designsummary["sensname"] == reference]["endreal1"])
        mask = real_mask(dfr_summed, startreal, endreal)
        ref_avg = dfr_summed[mask][response].mean()
    elif reference.isdigit():
        if int(reference) >= 0 and int(reference) <= dfr_summed.REAL.max():
            startreal = int(reference)
            endreal = int(reference)
            mask = real_mask(dfr_summed, startreal, endreal)
            ref_avg = dfr_summed[mask][response].mean()
    else:
        raise ValueError("Reference should be a sensname or a real number", reference)

    # for each sensitivity calculate statistics
    for sensno in range(len(designsummary)):
        sensname = designsummary.loc[sensno]["sensname"]
        startreal = designsummary.loc[sensno]["startreal1"]
        endreal = designsummary.loc[sensno]["endreal1"]
        mask = real_mask(dfr_summed, startreal, endreal)
        numreal1 = len(dfr_summed.REAL[mask])
        if numreal1 > 0:
            avg1 = dfr_summed[mask][response].mean() - ref_avg
        else:
            avg1 = 0
            print(
                "Warning: Number of ok realizations is 0 in"
                "sensitivity {} case1".format(sensname)
            )

        if designsummary.loc[sensno]["senstype"] == "mc":
            if numreal1 > 0:
                p90 = dfr_summed[mask][response].quantile(0.10) - ref_avg
                p10 = dfr_summed[mask][response].quantile(0.90) - ref_avg
            else:
                p90 = 0
                p10 = 0
                print(
                    "Warning: Number of ok realizations is 0 in"
                    "sensitivity {}".format(sensname)
                )
            subset1name = "p90"
            subset2name = "p10"
            tornadoinput.loc[sensno] = [
                sensno,
                sensname,
                p90,
                p10,
                subset1name,
                subset2name,
                numreal1,
                numreal1,
            ]
        elif designsummary.loc[sensno]["senstype"] == "scalar":
            subset1name = designsummary.loc[sensno]["casename1"]
            # if case 2 exists
            if designsummary.loc[sensno]["casename2"] is not None:
                startreal = designsummary.loc[sensno]["startreal2"]
                endreal = designsummary.loc[sensno]["endreal2"]
                mask = real_mask(dfr_summed, startreal, endreal)
                numreal2 = len(dfr_summed.REAL[mask])
                if numreal2 > 0:
                    avg2 = dfr_summed[mask][response].mean() - ref_avg
                else:
                    avg2 = 0
                    print(
                        "Warning: Number of ok realizations is 0 in"
                        "sensitivity {} case2".format(sensname)
                    )
                subset2name = designsummary.loc[sensno]["casename2"]
            else:
                avg2 = 0
                numreal2 = 0
                subset2name = None
            # If case1 has a higher value than case 2, swap values and names
            if avg2 < avg1:
                avg1, avg2 = avg2, avg1
                numreal1, numreal2 = numreal2, numreal1
                subset1name, subset2name = subset2name, subset1name
            tornadoinput.loc[sensno] = [
                sensno,
                sensname,
                avg1,
                avg2,
                subset1name,
                subset2name,
                numreal1,
                numreal2,
            ]
        elif designsummary.loc[sensno]["senstype"] == "ref":
            continue
        else:
            raise ValueError(
                "Sensitivity type should be 'mc' or 'scalar': "
                + designsummary.loc[sensno]["senstype"]
                + "\n"
                + "Something wrong with designsummary?"
            )

    tornadoinput["true_low"] = (tornadoinput["low"] + ref_avg).round(2)
    tornadoinput["true_high"] = (tornadoinput["high"] + ref_avg).round(2)

    if scale == "percentage":
        if ref_avg != 0:
            tornadoinput["low"] = 100 * tornadoinput["low"] / ref_avg
            tornadoinput["high"] = 100 * tornadoinput["high"] / ref_avg
        else:
            tornadoinput["low"] = 0
            tornadoinput["high"] = 0

    # Cut decimals for display
    tornadoinput["low"] = tornadoinput["low"].round(2)
    tornadoinput["high"] = tornadoinput["high"].round(2)

    # Drops sensitivities smaller than reference if specified
    if cutbyref and tornadoinput["sensname"].str.contains(reference).any():
        tornadoinput = cut_by_ref(tornadoinput, reference)

    # Sorts so that largest sensitivities comes first
    if sortsens:
        tornadoinput = sort_by_max(tornadoinput)

    # Return results that can be used for plotting
    tornadoinput = tornadoinput.drop(["sensno"], axis=1).set_index("sensname")

    return (tornadoinput, ref_avg)



"""Module for random sampling of parameter values from
distributions. For use in generation of design matrices
"""
from math import exp
import re
import numpy
import numpy.linalg as la
import scipy.stats
import pandas as pd


def _check_dist_params_normal(dist_params):
    if len(dist_params) != 2 and len(dist_params) != 4:
        status = False
        msg = (
            "Normal distribution must have 2 parameters"
            " or 4 for a truncated normal, "
            "but had " + str(len(dist_params)) + " parameters. "
        )
    elif not all(is_number(param) for param in dist_params):
        status = False
        msg = "Parameters for normal distribution must be numbers. "
    elif float(dist_params[1]) < 0:
        status = False
        msg = "Stddev for normal distribution must be >= 0. "
    else:
        status = True
        msg = ""

    return status, msg


def _check_dist_params_lognormal(dist_params):
    if len(dist_params) != 2:
        status = False
        msg = (
            "Lognormal distribution must have 2 parameters, "
            "but had " + str(len(dist_params)) + " parameters. "
        )
    elif not (is_number(dist_params[0]) and is_number(dist_params[1])):
        status = False
        msg = "Parameters for lognormal distribution must be numbers. "
    elif float(dist_params[1]) < 0:
        status = False
        msg = "Lognormal distribution must have stddev >= 0. "
    else:
        status = True
        msg = ""

    return status, msg


def _check_dist_params_uniform(dist_params):
    if len(dist_params) != 2:
        status = False
        msg = (
            "Uniform distribution must have 2 parameters, "
            "but had " + str(len(dist_params)) + " parameters. "
        )
    elif not (is_number(dist_params[0]) and is_number(dist_params[1])):
        status = False
        msg = "Parameters for uniform distribution must be numbers. "
    elif float(dist_params[1]) < float(dist_params[0]):
        status = False
        msg = "Uniform distribution must have dist_param2 >= dist_param1"
    else:
        status = True
        msg = ""

    return status, msg


def _check_dist_params_triang(dist_params):
    if len(dist_params) != 3:
        status = False
        msg = (
            "Triangular distribution must have 3 parameters, "
            "but had " + str(len(dist_params)) + " parameters. "
        )
    elif not all(is_number(param) for param in dist_params):
        status = False
        msg = "Parameters for triangular distribution must be numbers. "
    elif not (
        (float(dist_params[2]) >= float(dist_params[1]))
        and (float(dist_params[1]) >= float(dist_params[0]))
    ):
        status = False
        msg = "Triangular distribution must have: low <= mode <= high. "
    else:
        status = True
        msg = ""

    return status, msg


def _check_dist_params_pert(dist_params):
    if len(dist_params) not in [3, 4]:
        status = False
        msg = (
            "pert distribution must have 3 or 4 parameters, "
            "but had " + str(len(dist_params)) + " parameters. "
        )
    elif not all(is_number(param) for param in dist_params):
        status = False
        msg = "Parameters for pert distribution must be numbers. "
    elif not (
        (float(dist_params[2]) >= float(dist_params[1]))
        and (float(dist_params[1]) >= float(dist_params[0]))
    ):
        status = False
        msg = "Pert distribution must have: low <= mode <= high. "
    else:
        status = True
        msg = ""

    return status, msg


def _check_dist_params_logunif(dist_params):
    if len(dist_params) != 2:
        status = False
        msg = (
            "Log uniform distribution must have 2 parameters. "
            "but had " + str(len(dist_params)) + " parameters. "
        )
    elif not (is_number(dist_params[0]) and is_number(dist_params[1])):
        status = False
        msg = "Parameters for log uniform distribution must be numbers. "
    elif not (
        (float(dist_params[0]) > 0) and (float(dist_params[1]) >= float(dist_params[0]))
    ):
        status = False
        msg = "loguniform distribution must have low > 0 and high >=low. "
    else:
        status = True
        msg = ""

    return status, msg


def draw_values_normal(dist_parameters, numreals, normalscoresamples=None):
    """Draws values from normal distribution.

    Args:
        dist_parameters(list): [mean, std dev, min, max],
            min/max defining truncated normal
        numreals(int): number of realisations to draw
        normalscoresamples(list): samples for correlated parameters

    Returns:
        list of values
    """

    distribution = None
    if len(dist_parameters) == 2:  # normal
        status, msg = _check_dist_params_normal(dist_parameters)
        if status:
            dist_mean = float(dist_parameters[0])
            dist_stddev = float(dist_parameters[1])
            if normalscoresamples is not None:
                values = scipy.stats.norm.ppf(
                    scipy.stats.norm.cdf(normalscoresamples),
                    loc=dist_mean,
                    scale=dist_stddev,
                )
            else:
                distribution = scipy.stats.norm(dist_mean, dist_stddev)
                values = distribution.rvs(size=numreals)
        else:
            raise ValueError(msg)

    else:  # truncated normal or invalid
        status, msg = _check_dist_params_normal(dist_parameters)
        if status:
            mean = float(dist_parameters[0])
            sigma = float(dist_parameters[1])
            clip1 = float(dist_parameters[2])
            clip2 = float(dist_parameters[3])
            low = (clip1 - mean) / sigma
            high = (clip2 - mean) / sigma
            if normalscoresamples is not None:
                values = scipy.stats.truncnorm.ppf(
                    scipy.stats.norm.cdf(normalscoresamples),
                    low,
                    high,
                    loc=mean,
                    scale=sigma,
                )
            else:
                distribution = scipy.stats.truncnorm(low, high, loc=mean, scale=sigma)
                values = distribution.rvs(size=numreals)
        else:
            raise ValueError(msg)
    return values


def draw_values_lognormal(dist_parameters, numreals, normalscoresamples=None):
    """Draws values from lognormal distribution.

    Args:
        dist_parameters(list): [mu, sigma] for the logarithm of the variable
        numreals(int): number of realisations to draw
        normalscoresamples(list): samples for correlated parameters

    Returns:
        list of values
    """
    distribution = None
    status, msg = _check_dist_params_lognormal(dist_parameters)
    if status:
        mean = float(dist_parameters[0])
        sigma = float(dist_parameters[1])
        if normalscoresamples is not None:
            values = scipy.stats.lognorm.ppf(
                scipy.stats.norm.cdf(normalscoresamples),
                s=sigma,
                loc=0,
                scale=exp(mean),
            )
        else:
            distribution = scipy.stats.lognorm(s=sigma, scale=exp(mean))
            values = distribution.rvs(size=numreals)
    else:
        raise ValueError(msg)
    return values


def draw_values_uniform(dist_parameters, numreals, normalscoresamples=None):
    """Draws values from uniform distribution.

    Args:
        dist_parameters(list): [minimum, maximum]
        numreals(int): number of realisations to draw
        normalscoresamples(list): samples for correlated parameters

    Returns:
        list of values
    """
    distribution = None
    status, msg = _check_dist_params_uniform(dist_parameters)
    if status:
        low = float(dist_parameters[0])
        high = float(dist_parameters[1])
        uscale = high - low
        if normalscoresamples is not None:
            values = scipy.stats.uniform.ppf(
                scipy.stats.norm.cdf(normalscoresamples), loc=low, scale=uscale
            )

        else:
            distribution = scipy.stats.uniform(loc=low, scale=uscale)
            values = distribution.rvs(size=numreals)
    else:
        raise ValueError(msg)
    return values


def draw_values_triangular(dist_parameters, numreals, normalscoresamples=None):
    """Draws values from triangular distribution.

    Args:
        dist_parameters(list): [min, mode,  max]
        numreals(int): number of realisations to draw
        normalscoresamples(list): samples for correlated parameters

    Returns:
        list of values
    """
    status, msg = _check_dist_params_triang(dist_parameters)
    if status:
        low = float(dist_parameters[0])
        mode = float(dist_parameters[1])
        high = float(dist_parameters[2])
        if normalscoresamples is not None:
            if high == low:  # collapsed distribution
                print(
                    "Low and high parameters for triangular distribution"
                    " are equal. Using constant {}".format(low)
                )
                values = scipy.stats.uniform.ppf(
                    scipy.stats.norm.cdf(normalscoresamples), loc=low, scale=0
                )
            else:
                dist_scale = high - low
                shape = (mode - low) / dist_scale
                values = scipy.stats.triang.ppf(
                    scipy.stats.norm.cdf(normalscoresamples),
                    shape,
                    loc=low,
                    scale=dist_scale,
                )
        else:
            if high == low:  # collapsed distribution
                print(
                    "Low and high parameters for triangular distribution"
                    " are equal. Using constant {}".format(low)
                )
                distribution = scipy.stats.uniform(loc=low, scale=0)
                values = distribution.rvs(size=numreals)
            else:
                dist_scale = high - low
                shape = (mode - low) / dist_scale
                distribution = scipy.stats.triang(shape, loc=low, scale=dist_scale)
                values = distribution.rvs(size=numreals)
    else:
        raise ValueError(msg)
    return values


def draw_values_pert(dist_parameters, numreals, normalscoresamples=None):
    """Draws values from pert distribution.

    Args:
        dist_parameters(list): [min, mode,  max, scale]
            where scale is only specified
            for a 4 parameter pert distribution
        numreals(int): number of realisations to draw
        normalscoresamples(list): samples for correlated parameters

    Returns:
        list of values
    """

    status, msg = _check_dist_params_pert(dist_parameters)
    if status:
        low = float(dist_parameters[0])
        mode = float(dist_parameters[1])
        high = float(dist_parameters[2])
        if len(dist_parameters) == 4:
            scale = float(dist_parameters[3])
        else:
            scale = 4  # pert 3 parameter distribution

        if normalscoresamples is not None:
            if high == low:  # collapsed distribution
                print(
                    "Low and high parameters for pert distribution"
                    " are equal. Using constant {}".format(low)
                )
                values = scipy.stats.uniform.ppf(
                    scipy.stats.norm.cdf(normalscoresamples), loc=low, scale=0
                )
            else:
                muval = (low + high + scale * mode) / (scale + 2)
                if muval == mode:
                    alpha1 = (scale / 2) + 1
                else:
                    alpha1 = ((muval - low) * (2 * mode - low - high)) / (
                        (mode - muval) * (high - low)
                    )

                alpha2 = alpha1 * (high - muval) / (muval - low)
                values = scipy.stats.beta.ppf(
                    scipy.stats.norm.cdf(normalscoresamples), alpha1, alpha2
                )
        else:

            if high == low:  # collapsed distribution
                print(
                    "Low and high parameters for pert distribution"
                    " are equal. Using constant {}".format(low)
                )
                distribution = scipy.stats.uniform(loc=low, scale=0)
            else:
                muval = (low + high + scale * mode) / (scale + 2)
                if muval == mode:
                    alpha1 = (scale / 2) + 1
                else:
                    alpha1 = ((muval - low) * (2 * mode - low - high)) / (
                        (mode - muval) * (high - low)
                    )

                alpha2 = alpha1 * (high - muval) / (muval - low)
                distribution = scipy.stats.beta(alpha1, alpha2)
            values = distribution.rvs(size=numreals)

    else:
        raise ValueError(msg)
    # For pert distribution scale afterwards:
    values = values * (dist_parameters[2] - dist_parameters[0]) + dist_parameters[0]

    return values


def draw_values_loguniform(dist_parameters, numreals, normalscoresamples=None):
    """Draws values from loguniform distribution.

    Args:
        dist_parameters(list): [minimum, maximum]
        numreals(int): number of realisations to draw
        normalscoresamples(list): samples for correlated parameters

    Returns:
        list of values
    """

    status, msg = _check_dist_params_logunif(dist_parameters)
    if status:
        low = float(dist_parameters[0])
        high = float(dist_parameters[1])
        if normalscoresamples is not None:
            values = scipy.stats.reciprocal.ppf(
                scipy.stats.norm.cdf(normalscoresamples), low, high
            )
        else:
            distribution = scipy.stats.reciprocal(low, high)
            values = distribution.rvs(size=numreals)
    else:
        raise ValueError(msg)

    return values


def draw_values(distname, dist_parameters, numreals, normalscoresamples=None):
    """
    Prepare scipy distributions with parameters

    Args:
        distname (str): distribution name 'normal', 'lognormal', 'triang',
                        'uniform', 'logunif', 'discrete', 'pert'
        dist_parameters (list): list with parameters for distribution

    Returns:
        scipy.stats distribution with parameters
    """
    if distname[0:4].lower() == "norm":
        values = draw_values_normal(dist_parameters, numreals, normalscoresamples)

    elif distname[0:4].lower() == "logn":
        values = draw_values_lognormal(dist_parameters, numreals, normalscoresamples)

    elif distname[0:4].lower() == "unif":
        values = draw_values_uniform(dist_parameters, numreals, normalscoresamples)

    elif distname[0:6].lower() == "triang":
        values = draw_values_triangular(dist_parameters, numreals, normalscoresamples)

    elif distname[0:4].lower() == "pert":
        values = draw_values_pert(dist_parameters, numreals, normalscoresamples)

    elif distname[0:7].lower() == "logunif":
        values = draw_values_loguniform(dist_parameters, numreals, normalscoresamples)

    elif distname[0:5].lower() == "const":
        if normalscoresamples is not None:
            raise ValueError(
                "Parameter with const distribution "
                "was defined in correlation matrix "
                "but const distribution cannot "
                "be used with correlation. "
            )
        values = [dist_parameters[0]] * numreals

    elif distname[0:4].lower() == "disc":
        if normalscoresamples is not None:
            raise ValueError(
                "Parameter with discrete distribution "
                "was defined in correlation matrix, "
                "but discrete distribution cannot "
                "be used with correlation. "
            )
        status, result = sample_discrete(dist_parameters, numreals)
        if status:
            values = result
        else:
            raise ValueError(result)
    else:
        raise ValueError("distribution name {} is not implemented".format(distname))

    return values


def sample_discrete(dist_params, numreals):
    """Sample from discrete distribution

    Args:
        dist_params(list): parameters for distribution
            dist_params[0] is possible outcomes separated
            by comma
            dist_params[1] is probabilities for each outcome,
            separated by comma
        numreals (int): number of realisations to draw

    Returns:
        np.ndarray: values drawn from distribution
    """
    status = True
    outcomes = re.split(",", dist_params[0])
    outcomes = [item.strip() for item in outcomes]
    if len(dist_params) == 2:  # non uniform
        weights = re.split(",", dist_params[1])
        if len(outcomes) != len(weights):
            raise ValueError(
                "Number of weights for discrete distribution "
                "is not the same as number of values."
            )
        weightnmbr = [float(weight) for weight in weights]
        fractions = [weight / sum(weightnmbr) for weight in weightnmbr]
        values = numpy.random.choice(outcomes, numreals, p=fractions)
    elif len(dist_params) == 1:  # uniform
        values = numpy.random.choice(outcomes, numreals)
    else:
        status = False
        values = "Wrong input for discrete distribution"

    return status, values


def is_number(teststring):
    """Test if a string can be parsed as a float"""
    try:
        if not numpy.isnan(float(teststring)):
            return True
        return False  # It was a "number", but it was NaN.
    except ValueError:
        return False


def read_correlations(corr_dict, corrsheet):
    """Reading correlation info for a
    monte carlo sensitivity

    Args:
        corr_dict (OrderedDict): correlation info

    Returns:
        pd.DataFrame: Dataframe with correlations, parameter names
            as column and index
    """
    correlations = None
    filename = corr_dict["inputfile"]
    if corrsheet in corr_dict["sheetnames"]:
        if str(filename).endswith(".xlsx"):
            correlations = pd.read_excel(
                filename, corrsheet, index_col=0, engine="openpyxl"
            )
            correlations.dropna(axis=0, how="all", inplace=True)
            correlations = correlations.loc[
                :, ~correlations.columns.str.contains("^Unnamed")
            ]
        else:
            raise ValueError(
                "Correlation matrix filename should be on "
                "Excel format and end with .xlsx "
            )
    else:
        raise ValueError(
            "Corr_sheet {} specified but cannot be "
            "found in list of sheetnames".format(corrsheet)
        )

    return correlations


def make_covariance_matrix(df_correlations, stddevs=None):
    """Read a Pandas DataFrame defining correlation coefficients for
    a set of multivariate normally distributed parameters, and build
    covariance matrix.

    The diagonal of the correlation coefficients matrix should be all
    ones. Variances are combined with the correlation coefficients to
    compute the covariance matrix.

    If the correlation matrix is not symmetric positive definite (SDP),
    the matrix is projected onto the SDP manifold and returned (together
    with a warning). The algorithm is according to Higham (2000)

    Args:
        df_correlations (pd.DataFrame): correlation coefficients where
            columns and index are both parameter names. All parameter
            names in keys muvalst also exist in index and vice versa.

    Returns:
        covariance matrix
    """

    corr_matrix = numpy.array(df_correlations.values)

    # Assume upper triangular is empty, fill it:
    i_upper = numpy.triu_indices(len(df_correlations.columns), 1)
    corr_matrix[i_upper] = corr_matrix.T[i_upper]

    # Project to nearest symmetric positive definite matrix
    if not _is_positive_definite(corr_matrix):
        input_corr_matrix = corr_matrix.copy()
        corr_matrix = _nearest_positive_definite(corr_matrix)
        print("Input correlation matrix: ")
        with numpy.printoptions(precision=3, suppress=True):
            print(input_corr_matrix)
        print("Used closest positive semi-definite correlation matrix:")
        with numpy.printoptions(precision=3, suppress=True):
            print(corr_matrix)
    # Previously negative eigenvalues are now close to zero,
    # but might still be negative, that can be ignored

    # Support unity standard devistions
    if not stddevs:
        stddevs = len(corr_matrix) * [1]

    # Now generate the covariance matrix
    dim = len(stddevs)
    diag = numpy.identity(dim)
    diag[range(dim), range(dim)] = stddevs
    cov_matrix = numpy.dot(diag, corr_matrix)
    cov_matrix = numpy.dot(cov_matrix, diag)

    return cov_matrix


def _nearest_positive_definite(a_mat):
    """Implementation taken from:
    https://stackoverflow.com/questions/43238173/
    python-convert-matrix-to-positive-semi-definite/43244194#43244194
    """

    b_mat = (a_mat + a_mat.T) / 2
    _, s_mat, v_mat = la.svd(b_mat)

    h_mat = numpy.dot(v_mat.T, numpy.dot(numpy.diag(s_mat), v_mat))

    a2_mat = (b_mat + h_mat) / 2

    a3_mat = (a2_mat + a2_mat.T) / 2

    if _is_positive_definite(a3_mat):
        return a3_mat

    spacing = numpy.spacing(la.norm(a_mat))
    identity = numpy.eye(a_mat.shape[0])
    kiter = 1
    while not _is_positive_definite(a3_mat):
        mineig = numpy.min(numpy.real(la.eigvals(a3_mat)))
        a3_mat += identity * (-mineig * kiter ** 2 + spacing)
        kiter += 1

    return a3_mat


def _is_positive_definite(b_mat):
    """Returns true when input is positive-definite, via Cholesky"""
    try:
        _ = la.cholesky(b_mat)
        return True
    except la.LinAlgError:
        return False



"""
The qcforward methods module

This is a function based approach, but in many cases it may be better
if the user make an explicit instance in the calling script, in particular if
the job is about to be read numerous times with the ``reuse`` option


I.e::

    from fmu.tools import qcforward

    # .. define data

    qcforward.wellzonation_vs_grid(data)

    # vs

    qcjob = qcforward.WellZonationVsGrid()
    qcjob.run(data)


"""
from fmu.tools.qcforward._blockedwells_vs_gridprops import BlockedWellsVsGridProperties
from fmu.tools.qcforward._grid_quality import GridQuality
from fmu.tools.qcforward._grid_statistics import GridStatistics
from fmu.tools.qcforward._wellzonation_vs_grid import WellZonationVsGrid


def wellzonation_vs_grid(data, project=None):
    """Check well zonation or perforations vs 3D grid.

    Args:
        data (dict): This is dictonary telling where data comes from

    """

    wzong = WellZonationVsGrid()
    wzong.run(data, project=project)


def grid_statistics(data, project=None):
    """Check statistics in 3D grid against user input.

    Args:
        data (dict or str): The input data either as a Python dictionary or
            a path to a YAML file
    """

    gps = GridStatistics()
    gps.run(data, project=project)


def blockedwells_vs_gridproperties(data, project=None):
    """Check blocked wells vs vs grid props.

    Args:
        data (dict or str): The input data either as a Python dictionary or
            a path to a YAML file
    """

    bwgp = BlockedWellsVsGridProperties()
    bwgp.run(data, project=project)


def grid_quality(data, project=None):
    """Check grid quality in 3D grid against user input.

    Args:
        data (dict): The input data either as a Python dictionary or
            a path to a YAML file
    """

    gqual = GridQuality()
    gqual.run(data, project=project)



"""Test code for RMS volumetrics parsing"""

import sys
import subprocess
from pathlib import Path

import pytest

import pandas as pd

from fmu.tools.rms import volumetrics


TESTDIR = Path(__file__).parent / "volumetricsdata"


@pytest.mark.parametrize(
    "multiline_str, filename, expected_df",
    [
        pytest.param(
            "",
            "foo.txt",
            None,
            marks=pytest.mark.xfail(raises=pd.errors.EmptyDataError),
        ),
        pytest.param(
            "foo bar",
            "foo.txt",
            None,
            marks=pytest.mark.xfail(raises=pd.errors.EmptyDataError),
        ),
        pytest.param(
            "Zone Bulk\nUpper 1",
            "foo.txt",
            pd.DataFrame([{"BULK": 1}]),
            marks=pytest.mark.xfail(
                raises=ValueError, reason="Not able to guess phase"
            ),
        ),
        pytest.param(
            "Zone Bulk\nUpper 1",
            "oil.txt",
            # This "fails" as there was not a double space between Zone and Bulk
            pd.DataFrame([{"Zone Bulk": "Upper 1"}]),
        ),
        (
            # Integer zone, returned as string datatype
            "Zone  Bulk\n1  1.0",
            "oil.txt",
            pd.DataFrame([{"ZONE": "1", "BULK_OIL": 1.0}]),
        ),
        (
            # Integer Region, returned as string datatype
            "Region index  Bulk\n1  1.0",
            "oil.txt",
            pd.DataFrame([{"REGION": "1", "BULK_OIL": 1.0}]),
        ),
        (
            # Floating point Region, returned as string datatype.
            # User is probably up for trouble if this is used.
            "Region index  Bulk\n1.0  1.0",
            "oil.txt",
            pd.DataFrame([{"REGION": "1.0", "BULK_OIL": 1.0}]),
        ),
        (
            # Two spaces:
            "Zone  Bulk\nUpper  1.0",
            "oil.txt",
            pd.DataFrame([{"ZONE": "Upper", "BULK_OIL": 1.0}]),
        ),
        (
            # Three spaces:
            "Zone   Bulk\nUpper  1.0",
            "oil.txt",
            pd.DataFrame([{"ZONE": "Upper", "BULK_OIL": 1.0}]),
        ),
        (
            # Tabs don't work:
            "Zone\tBulk\nUpper\t1.0",
            "oil.txt",
            pd.DataFrame([{"Zone\tBulk": "Upper\t1.0"}]),
        ),
        (
            "Zone  Bulk\nUpper  1.0",
            "gas.txt",
            pd.DataFrame([{"ZONE": "Upper", "BULK_GAS": 1.0}]),
        ),
        (
            # All supported columns:
            "Zone  Region index  Facies  License boundaries  "
            "Bulk  Net  Hcpv  Pore  Stoiip  Assoc.Gas\n"
            "Upper  West  GoodSand  NO  2  1  1  1  1  0.1",
            "oil.txt",
            pd.DataFrame(
                [
                    {
                        "ZONE": "Upper",
                        "REGION": "West",
                        "FACIES": "GoodSand",
                        "LICENSE": "NO",
                        "BULK_OIL": 2,
                        "NET_OIL": 1,
                        "HCPV_OIL": 1,
                        "PORV_OIL": 1,
                        "STOIIP_OIL": 1,
                        "ASSOCIATEDGAS_OIL": 0.1,
                    }
                ]
            ),
        ),
        (
            # All supported columns:
            "Zone  Region index  Facies  License boundaries  "
            "Bulk  Net  Hcpv  Pore  Giip  Assoc.Liquid\n"
            "Upper  West  GoodSand  NO  2  1  1  1  1  0.1",
            "gas.txt",
            pd.DataFrame(
                [
                    {
                        "ZONE": "Upper",
                        "REGION": "West",
                        "FACIES": "GoodSand",
                        "LICENSE": "NO",
                        "BULK_GAS": 2,
                        "NET_GAS": 1,
                        "HCPV_GAS": 1,
                        "PORV_GAS": 1,
                        "GIIP_GAS": 1,
                        "ASSOCIATEDOIL_GAS": 0.1,
                    }
                ]
            ),
        ),
    ],
)
def test_rms_to_volumetrics(multiline_str, filename, expected_df, tmpdir):
    tmpdir.chdir()
    Path(filename).write_text(multiline_str)
    pd.testing.assert_frame_equal(
        volumetrics.rmsvolumetrics_txt2df(filename), expected_df, check_like=True
    )


def test_volumetrics():
    """Test parsing of many real examples from RMS"""

    for filename in TESTDIR.glob("*.txt"):
        dframe = volumetrics.rmsvolumetrics_txt2df(filename)

        # Check that we did get some data:
        assert len(dframe) > 0

        # Check for no non-uppercase column names:
        collist = list(dframe.columns)
        assert [x.upper() for x in collist] == collist

        # Check for no string 'Totals' in any cell
        assert "Totals" not in dframe.to_string()

    # Test zone renamer:
    def myrenamer(a_zone):
        """Callback function for zone renaming"""
        return a_zone.replace("Larsson", "E")

    dframe = volumetrics.rmsvolumetrics_txt2df(
        TESTDIR / "14_geo_gas_1.txt", zonerenamer=myrenamer
    )
    assert "Larsson" not in dframe.to_string()
    assert "E3_1" in dframe.to_string()

    # Test columnrenamer:
    columnrenamer = {"Region index": "FAULTSEGMENT"}  # this will override
    dframe = volumetrics.rmsvolumetrics_txt2df(
        TESTDIR / "14_geo_gas_1.txt",
        columnrenamer=columnrenamer,
    )
    assert "FAULTSEGMENT" in dframe.columns


def test_merge_rms_volumetrics_explicit(tmpdir):
    """Test finding and merging of multiple RMS volumetrics files"""
    tmpdir.chdir()
    Path("test_oil_1.txt").write_text("Zone  Bulk\nUpper  1.0")
    Path("test_gas_1.txt").write_text("Zone  Bulk\nUpper  1.0")
    Path("test_total_1.txt").write_text("Zone  Bulk\nUpper  1.0")

    expected_frame = pd.DataFrame(
        [{"ZONE": "Upper", "BULK_OIL": 1.0, "BULK_GAS": 1.0, "BULK_TOTAL": 1.0}]
    )
    pd.testing.assert_frame_equal(
        volumetrics.merge_rms_volumetrics("test"),
        expected_frame,
        check_like=True,
    )

    # Add a dummy file that we should not discover, and thus not interfere:
    Path("test_dummy_1.txt").write_text("aslkdjfa")
    pd.testing.assert_frame_equal(
        volumetrics.merge_rms_volumetrics("test"),
        expected_frame,
        check_like=True,
    )

    # What if there is no files found:
    with pytest.raises(OSError, match="No volumetrics files found for filebase foo"):
        volumetrics.merge_rms_volumetrics("foo")


def test_merge_rms_volumetrics_mixedcols(tmpdir):
    """Bulk in one file and Pore in another, just means we don't get all columns"""
    tmpdir.chdir()
    Path("test_oil_1.txt").write_text("Zone  Bulk\nUpper  1.0")
    Path("test_gas_1.txt").write_text("Zone  Pore\nUpper  1.0")
    Path("test_total_1.txt").write_text("Zone  Bulk\nUpper  1.0")

    expected_frame = pd.DataFrame(
        [{"ZONE": "Upper", "BULK_OIL": 1.0, "PORV_GAS": 1.0, "BULK_TOTAL": 1.0}]
    )
    pd.testing.assert_frame_equal(
        volumetrics.merge_rms_volumetrics("test"),
        expected_frame,
        check_like=True,
    )


@pytest.mark.integration
def test_commandlineclient_installed():
    """Test endpoint is installed"""
    assert subprocess.check_output(["rmsvolumetrics2csv", "-h"])


@pytest.mark.integration
def test_commandlineclient(tmpdir):
    """Test endpoint"""

    tmpdir.chdir()

    sys.argv = [
        "rmsvolumetrics2csv",
        str(TESTDIR / "1_geogrid_vol_oil_1.txt"),
        "--output",
        "geogrid_oil.csv",
    ]
    volumetrics.rmsvolumetrics2csv_main()

    disk_df = pd.read_csv("geogrid_oil.csv")
    assert "STOIIP_OIL" in disk_df

    sys.argv = [
        "rmsvolumetrics2csv",
        str(TESTDIR / "14_geo_gas_1.txt"),
        "--output",
        "geo_gas.csv",
    ]
    volumetrics.rmsvolumetrics2csv_main()
    disk_df = pd.read_csv("geo_gas.csv")
    assert "GIIP_GAS" in disk_df

    # Test that --phase will override the inferral

    sys.argv = [
        "rmsvolumetrics2csv",
        str(TESTDIR / "14_geo_gas_1.txt"),
        "--output",
        "geo_foophase.csv",
        "--phase",
        "FOOBAR",
    ]
    volumetrics.rmsvolumetrics2csv_main()
    disk_df = pd.read_csv("geo_foophase.csv")
    assert "GIIP_FOOBAR" in disk_df
    # Test that parent directories will be created for output
    sys.argv = [
        "rmsvolumetrics2csv",
        str(TESTDIR / "14_geo_gas_1.txt"),
        "--output",
        "foo/bar/com.csv",
    ]
    volumetrics.rmsvolumetrics2csv_main()
    disk_df = pd.read_csv("foo/bar/com.csv")
    assert "GIIP_GAS" in disk_df



from typing import List, Dict, Union, Any
import warnings

try:
    import _roxar  # type: ignore
except ModuleNotFoundError:
    warnings.warn("This script only supports interactive RMS usage", UserWarning)


def _set_safe_value(
    project: Any, surf_type: str, name: str, data_type: str, value: float
):
    """Set the horizon or zone surface to the defined value.

    Args:
        project: the roxar project.
        surf_type: the type of surface.
            It should be the string "horizons" or "zones".
        name: the name the of surface.
        value: the value (int or float) to be assigned to the surface
    """

    try:
        if surf_type == "horizons":
            surf = project.horizons[name][data_type]
        else:
            surf = project.zones[name][data_type]
        grid2d = surf.get_grid()
        grid2d.set_values(grid2d.get_values() * 0.0 + value)
        surf.set_grid(grid2d)
        print(" >> >> " + name)
    except Exception as e:
        print(" >> >> " + name + " cannot be modified")
        print(e)


def _set_safe_empty(project: Any, surf_type: str, name: str, data_type: str):
    """Set empty the horizon or zone surface.

    Args:
        project: the roxar project.
        surf_type: the type of surface.
            It should be the string "horizons" or "zones".
        name: the name the of surface.
    """
    try:
        if surf_type == "horizons":
            surf = project.horizons[name][data_type]
        else:
            surf = project.zones[name][data_type]
        surf.set_empty()
        print(" >> >> " + name)
    except Exception as e:
        print(" >> >> " + name + " cannot be modified")
        print(e)


def _set_surfaces_value(
    project: Any,
    surf_type: str,
    dict_val: Union[List, Dict],
    value: float,
):
    """Set a group of surfaces to a given value.

    Args:
        project: roxar project.
        surf_type: the type of surface.
            It should be the string "horizon" or "zone".
        dict_val: a list of surfaces or a dictionary of surfaces categories
            (keys) with a list of surfaces names (value).
        value: the value to assign to the surfaces.
    """
    if surf_type == "horizons":
        surfaces = project.horizons
    elif surf_type == "zones":
        surfaces = project.zones
    else:
        raise ValueError("surf_type must be 'horizons' or 'zones'")

    if isinstance(dict_val, list):
        # work directly at horizon/zone category level
        for data_type in dict_val:
            print(" >> " + data_type)
            for surface in surfaces:
                _set_safe_value(project, surf_type, surface.name, data_type, value)
    elif isinstance(dict_val, dict):
        # check setup for each horizon/zone category (list vs. all)
        surf_cat = dict_val.keys()
        for data_type in surf_cat:
            print(" >> " + data_type)
            surf_names = dict_val[data_type]
            if isinstance(surf_names, str):
                if surf_names == "all":
                    for surface in surfaces:
                        _set_safe_value(
                            project, surf_type, surface.name, data_type, value
                        )
                else:
                    raise ValueError(
                        "keyword '" + surf_names + "' not recognized, 'all' expected!"
                    )
            elif isinstance(surf_names, list):
                for surf_name in surf_names:
                    _set_safe_value(project, surf_type, surf_name, data_type, value)
    else:
        raise TypeError(
            "Value associated with key '"
            + surf_type
            + "' must be of type list or dict!"
        )


def _set_surfaces_empty(project: Any, surf_type: str, dict_val: Union[List, Dict]):
    """Set empty a group of surfaces.

    Args:
        project: roxar project.
        surf_type: the type of surface.
            It should be the string "horizon" or "zone".
        dict_val: a list of surfaces or a dictionary of surfaces categories
            (keys) with a list of surfaces names (value).
    """
    if surf_type == "horizons":
        surfaces = project.horizons
    elif surf_type == "zones":
        surfaces = project.zones
    else:
        raise ValueError("surf_type must be 'horizons' or 'zones'")

    if isinstance(dict_val, list):
        # work directly at horizon/zone category level
        for data_type in dict_val:
            print(" >> " + data_type)
            for surface in surfaces:
                _set_safe_empty(project, surf_type, surface.name, data_type)
    elif isinstance(dict_val, dict):
        # check setup for each horizon/zone category (list vs. all)
        surf_cat = dict_val.keys()
        for data_type in surf_cat:
            print(" >> " + data_type)
            surf_names = dict_val[data_type]
            if isinstance(surf_names, str):
                if surf_names == "all":
                    for surface in surfaces:
                        _set_safe_empty(project, surf_type, surface.name, data_type)
                else:
                    raise ValueError(
                        "keyword '" + surf_names + "' not recognized, 'all' expected!"
                    )
            elif isinstance(surf_names, list):
                for surf_name in surf_names:
                    _set_safe_empty(project, surf_type, surf_name, data_type)
    else:
        raise TypeError(
            "Value associated with key '"
            + surf_type
            + "' must be of type list or dict!"
        )


def set_data_constant(config: Dict):
    """Set data from RMS constant.

    This method is a utility in order to set surface and 3D grid property data
    to a given value. The value must be of the correct type (if discrete 3D
    property for example). The purpose of it is to make sure that those data
    are properly generated by the modelling workflow and not inherited from a
    previous run with the corresponding jobs deactivated. The data are set to a
    value triggering attention and not deleted in order not to reset some jobs
    in RMS.

    The input of this method is a Python dictionary with defined keys. The keys
    "project" and "value" are required while "horizons", "zones" and
    "grid_models" are optional (at least one of them should be provided for the
    method to have any effect).

    Args:
        project: The roxar magic keyword ``project`` refering to the current
            RMS project.

        value: The constant value to assign to the data. It could be 0 or -999
            for example. If discrete properties from grid models are modified,
            the value should be applicable (integer).

        horizons: A Python dictionary where each key corresponds to the name of
            the horizons category where horizon data need to be modified. The
            value associated to this key should be a list of horizon names to
            modify. If a string ``all`` is assigned instead of a list, all
            available horizon names for this category will be used.
            Alternatively, if a list of horizons categories is given instead of
            a dictionary, the method will apply to all horizons within these
            horizons categories.

        zones: A Python dictionary where each key corresponds to the name of
            the zones category where zone data need to be modified. The value
            associated to this key should be a list of zone names to modify. If
            a string ``all`` is assigned instead of a list, all available zone
            names for this category will be used.
            Alternatively, if a list of zones categories is given instead of a
            dictionary, the method will apply to all zones within these zones
            categories.

        grid_models: A Python dictionary where each key corresponds to the name
            of the grid models where properties need to be modified. The
            value associated to this key should be a list of property names to
            modify. If a string ``all`` is assigned instead of a list, all
            available properties for this grid model name will be used.
            Alternatively, if a list of grid models names is given instead of a
            dictionary, the method will apply to all properties within these
            grid models.
    """
    if not isinstance(config, dict):
        raise TypeError("Argument must be a Python dictionary!")
    assert "project" in config.keys(), "Input dict must contain key 'project'!"
    project = config["project"]
    if not isinstance(project, _roxar.Project):
        raise RuntimeError("This run must be ran in an RoxAPI environment!")

    assert "value" in config.keys(), "Input dict must contain key 'value'!"
    value = config["value"]

    # HORIZON DATA
    if "horizons" in config.keys():
        print("Set horizons values to " + str(value) + "...")
        _set_surfaces_value(project, "horizons", config["horizons"], value)

    # ZONE DATA
    if "zones" in config.keys():
        print("Set zones values to " + str(value) + "...")
        _set_surfaces_value(project, "zones", config["zones"], value)

    # GRID MODEL DATA
    if "grid_models" in config.keys():
        print("Set 3D grid properties values to " + str(value) + "...")
        if isinstance(config["grid_models"], list):
            # work directly at grid models level
            for gridname in config["grid_models"]:
                print(" >> " + gridname)
                grid = project.grid_models[gridname]
                for prop in grid.properties:
                    try:
                        prop.set_values(prop.get_values() * 0 + value)
                        print(" >> >> " + prop.name)
                    except Exception as e:
                        print(" >> >> " + prop.name + " is already empty")
                        print(e)
        elif isinstance(config["grid_models"], dict):
            # check setup for each grid models (list vs. all)
            gridnames = config["grid_models"].keys()
            for gridname in gridnames:
                print(" >> " + gridname)
                grid = project.grid_models[gridname]
                propnames = config["grid_models"][gridname]
                if isinstance(propnames, str):
                    if propnames == "all":
                        for prop in grid.properties:
                            try:
                                prop.set_values(prop.get_values() * 0 + value)
                                print(" >> >> " + prop.name)
                            except Exception as e:
                                print(" >> >> " + prop.name + " is already empty")
                                print(e)
                    else:
                        raise Exception(
                            "keyword " + propnames + "not recognized, 'all' expected!"
                        )
                elif isinstance(propnames, list):
                    for propname in propnames:
                        try:
                            prop = grid.properties[propname]
                            prop.set_values(prop.get_values() * 0 + value)
                            print(" >> >> " + prop.name)
                        except Exception as e:
                            print(" >> >> " + prop.name + " is already empty")
                            print(e)
        else:
            raise TypeError(
                "Value associated with key 'zones' must be of type list or dict!"
            )

    print("End of function set_data_constant().")


def set_data_empty(config: Dict):
    """Set data from RMS empty.

    This method is a utility in order to set empty surface and 3D grid property
    data. The value must be of the correct type (if discrete 3D property for
    example). The purpose of it is to make sure that those data are properly
    generated by the modelling workflow and not inherited from a previous run
    with the corresponding jobs deactivated.

    The input of this method is a Python dictionary with defined keys. The keys
    "project" and "value" are required while "horizons", "zones" and
    "grid_models" are optional (at least one of them should be provided for the
    method to have any effect).

    Input configrations:

    project: The roxar magic keyword ``project`` refering to the current
        RMS project.

    horizons: A Python dictionary where each key corresponds to the name of
        the horizons category where horizon data need to be made empty. The
        value associated to this key should be a list of horizon names to
        modify. If a string ``all`` is assigned instead of a list, all
        available horizon names for this category will be used.
        Alternatively, if a list of horizons categories is given instead of
        a dictionary, the method will apply to all horizons within these
        horizons categories.

    zones: A Python dictionary where each key corresponds to the name of
        the zones category where zone data need to be made empty. The value
        associated to this key should be a list of zone names to modify. If
        a string ``all`` is assigned instead of a list, all available zone
        names for this category will be used.
        Alternatively, if a list of zones categories is given instead of a
        dictionary, the method will apply to all zones within these zones
        categories.

    grid_models: A Python dictionary where each key corresponds to the name
        of the grid models where properties need to be made empty. The
        value associated to this key should be a list of property names to
        modify. If a string ``all`` is assigned instead of a list, all
        available properties for this grid model name will be used.
        Alternatively, if a list of grid models names is given instead of a
        dictionary, the method will apply to all properties within these
        grid models.



    Args:

        config: Configration as a dictionary. See examples in documentation

    """

    if not isinstance(config, dict):
        raise TypeError("Argument must be a Python dictionary!")
    assert "project" in config.keys(), "Input dict must contain key 'project'!"
    project = config["project"]
    if not isinstance(project, _roxar.Project):
        raise RuntimeError("This run must be ran in an RoxAPI environment!")

    # HORIZON DATA
    if "horizons" in config.keys():
        print("Set empty horizons...")
        _set_surfaces_empty(project, "horizons", config["horizons"])

    # ZONE DATA
    if "zones" in config.keys():
        print("Set empty zones...")
        _set_surfaces_empty(project, "zones", config["zones"])

    # GRID MODEL DATA
    if "grid_models" in config.keys():
        print("Set empty 3D grid properties...")
        if isinstance(config["grid_models"], list):
            # work directly at grid models level
            for gridname in config["grid_models"]:
                print(" >> " + gridname)
                grid = project.grid_models[gridname]
                for prop in grid.properties:
                    prop.set_empty()
                    print(" >> >> " + prop.name)
        elif isinstance(config["grid_models"], dict):
            # check setup for each grid models (list vs. all)
            gridnames = config["grid_models"].keys()
            for gridname in gridnames:
                print(" >> " + gridname)
                grid = project.grid_models[gridname]
                propnames = config["grid_models"][gridname]
                if isinstance(propnames, str):
                    if propnames == "all":
                        for prop in grid.properties:
                            prop.set_empty()
                            print(" >> >> " + prop.name)
                    else:
                        raise Exception(
                            "keyword " + propnames + "not recognized, 'all' expected!"
                        )
                elif isinstance(propnames, list):
                    for propname in propnames:
                        prop = grid.properties[propname]
                        prop.set_empty()
                        print(" >> >> " + prop.name)
        else:
            raise TypeError(
                "Value associated with key 'zones' must be of type list or dict!"
            )

    print("End of function set_data_empty().")



"""Module for common utility functions"""

from itertools import combinations


def filter_df(dframe, filters):
    """Filter dataframe"""
    dframe = dframe.copy()
    for prop, filt in filters.items():

        if "include" in filt:
            if all(x in dframe[prop].unique() for x in filt["include"]):
                dframe = dframe[dframe[prop].isin(filt["include"])]
            else:
                raise ValueError(
                    f"One or more items in {filt['include']} "
                    f"does not exist in dataframe column {prop} "
                    f"Available values are: {dframe[prop].unique()}"
                )
        if "exclude" in filt:
            if all(x in dframe[prop].unique() for x in filt["exclude"]):
                dframe = dframe[~dframe[prop].isin(filt["exclude"])]
            else:
                raise ValueError(
                    f"One or more items in {filt['exclude']} "
                    f"does not exist in dataframe column {prop} "
                    f"Available values are: {dframe[prop].unique()}"
                )
        if "range" in filt:
            low_value, high_value = filt["range"]
            dframe = dframe[(dframe[prop] >= low_value) & (dframe[prop] <= high_value)]

    if dframe.empty:
        raise Exception("Empty dataframe - no data left after filtering")

    return dframe


def list_combinations(input_list):
    """Create a list of all possible combinations of an existing list"""
    combo_list = []
    for item in range(len(input_list), 0, -1):
        for combo in combinations(input_list, item):
            combo_list.append(list(combo))
    return combo_list



from pathlib import Path
from packaging import version

import pandas as pd

import pytest

from fmu.tools.sensitivities import (
    calc_tornadoinput,
    summarize_design,
)

TESTDATA = Path(__file__).parent / "data"


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_designsummary():
    """Test import and summary of design matrix"""

    snorrebergdesign = summarize_design(
        TESTDATA / "distributions/design.xlsx", "DesignSheet01"
    )
    # checking dimensions and some values in summary of design matrix
    assert snorrebergdesign.shape == (7, 9)
    assert snorrebergdesign["sensname"][0] == "rms_seed"
    assert snorrebergdesign["startreal2"][6] == 100
    assert snorrebergdesign["endreal2"][6] == 109
    assert snorrebergdesign["endreal1"].sum() == 333

    # Test same also when design matrix is in .csv format
    designcsv = summarize_design(TESTDATA / "distributions/design.csv")

    # checking dimensions and some values in summary of design matrix
    assert designcsv.shape == (7, 9)
    assert designcsv["sensname"][0] == "rms_seed"
    assert designcsv["startreal2"][6] == 100
    assert designcsv["endreal2"][6] == 109
    assert designcsv["endreal1"].sum() == 333


def test_calc_tornadoinput():
    """Test calculating values for tornadoplot input"""
    # Read file with summary of design
    summary = pd.read_csv(TESTDATA / "distributions/designsummary.csv", na_values="nan")
    des_summary = summary.where(pd.notnull(summary), None)

    # Read resultfile for the test ensemble
    results = pd.read_csv(TESTDATA / "results/geovolumes_collected.csv")

    # Calculate and check results of one tornado calculation
    (tornadotable, ref_value) = calc_tornadoinput(
        des_summary,
        results,
        "STOIIP_OIL",
        ["ITER", "ZONE", "REGION"],
        [["iter-0"], ["Nansen", "Larsson"], ["oil_zone_Nansen_Larsson"]],
        "rms_seed",
        "percentage",
    )

    assert int(tornadotable["low"].sum()) == -21
    assert int(tornadotable["high"].sum()) == 11
    assert int(ref_value) == 9330662

    # Check summing over all zones and regions before calculations
    (tornadotable, ref_value) = calc_tornadoinput(
        des_summary,
        results,
        "STOIIP_OIL",
        ["ITER", "ZONE", "REGION"],
        [["all"], ["all"], ["all"]],
        "rms_seed",
        "absolute",
    )

    assert int(tornadotable["low"].sum()) == -2142167
    assert int(tornadotable["high"].sum()) == 1478053
    assert int(ref_value) == 12855200



"""
This private module in qcforward is used for grid statistics
"""
from typing import Union, Dict
import collections
from pathlib import Path
import json
from jsonschema import validate
import pandas as pd

import fmu.tools
from fmu.tools.qcproperties.qcproperties import QCProperties

from fmu.tools._common import _QCCommon
from fmu.tools.qcforward._qcforward import QCForward


QCC = _QCCommon()


class _LocalData:
    def __init__(self):
        """Defining and hold data local for this routine"""
        self.actions = None
        self.nametag = None
        self.reportfile = None

    def parse_data(self, data):
        """Parsing the actual data"""
        self.nametag = data.get("nametag", "-")
        self.reportfile = data.get("report", None)
        self.actions = data["actions"]


class GridStatistics(QCForward):
    def run(
        self,
        data: Union[dict, str],
        project: Union[object, str] = None,
    ) -> None:
        """Main routine for evaulating if statistics from 3D grids is
        within user specified thresholds.

        The routine depends on existing fmu.tools functionality for
        extracting property statistics from 3D grids.

        Args:
            data (dict or str): The input data either as a Python dictionary or
                a path to a YAML file
            project (Union[object, str]): For usage inside RMS

        """

        self._data: dict = self.handle_data(data, project)
        # TO-DO:
        # self._validate_input(self._data)

        data = self._data
        QCC.verbosity = data.get("verbosity", 0)

        # parse data that are special for this check
        QCC.print_info("Parsing additional data...")
        self.ldata = _LocalData()
        self.ldata.parse_data(data)

        dfr = self.check_gridstatistics(project, data)
        QCC.print_debug(f"Results: \n{dfr}")

        self.evaluate_qcreport(dfr, "grid statistics")

    def check_gridstatistics(self, project, data):
        """
        Extract statistics per action and check if property value is
        within user specified limits.

        Returns a dataframe with results
        """

        qcp = QCProperties()

        results = []
        QCC.print_info("Checking status for items in actions...")
        for action in self.ldata.actions:
            # extract parameters from actions
            data_upd = self._extract_parameters_from_action(data, action)

            selectors, calculation = self._get_selecors_and_calculation(action)

            # Create datframe with statistics
            dframe = qcp.get_grid_statistics(project=project, data=data_upd)

            # Get value from statistics for given property and selectors
            value = self._get_statistical_value(
                dframe, action["property"], calculation, selectors
            )

            status = "OK"
            if "warn_outside" in action:
                if not action["warn_outside"][0] <= value <= action["warn_outside"][1]:
                    status = "WARN"
            if not action["stop_outside"][0] <= value <= action["stop_outside"][1]:
                status = "STOP"

            result = collections.OrderedDict()
            result["PROPERTY"] = action["property"]
            result["SELECTORS"] = f"{list(selectors.values())}"
            result["FILTERS"] = "yes" if "filters" in action else "no"
            result["CALCULATION"] = calculation
            result["VALUE"] = value
            result["STOP_LIMITS"] = f"{action['stop_outside']}"
            result["WARN_LIMITS"] = (
                f"{action['warn_outside']}" if "warn_outside" in action else "-"
            )
            result["STATUS"] = status
            result["DESCRIPTION"] = action.get("description", "-")

            results.append(result)

        dfr = self.make_report(
            results, reportfile=self.ldata.reportfile, nametag=self.ldata.nametag
        )

        return dfr

    @staticmethod
    def _validate_input(data: dict):
        """Validate data against JSON schemas"""

        # TODO: complete JSON files
        spath = Path(fmu.tools.__file__).parent / "qcforward" / "_schemas"

        schemafile = "grid_statistics_asfile.json"

        if "project" in data.keys():
            schemafile = "grid_statistics_asroxapi.json"

        with open((spath / schemafile), "r") as thisschema:
            schema = json.load(thisschema)

        validate(instance=data, schema=schema)

    @staticmethod
    def _extract_parameters_from_action(data: dict, action: Dict[str, dict]) -> dict:
        """
        Extract property and selector data from actions
        and convert to desired input format for QCProperties
        """
        data = data.copy()

        properties: dict = {}
        selectors: list = []
        filters: dict = {}

        if action["property"] not in properties:
            properties[action["property"]] = {"name": action["property"]}

        if "filters" in action:
            filters = action["filters"]

        if "selectors" in action:
            for prop, filt in action["selectors"].items():
                if prop not in selectors:
                    selectors.append(prop)
                filters[prop] = {"include": filt}

        data["properties"] = properties
        data["selectors"] = selectors
        data["filters"] = filters

        return data

    @staticmethod
    def _get_selecors_and_calculation(action: dict) -> tuple:
        """
        Get selectors and selected calculation from the action.
        If a discrete property has been input it is added to the selctors.
        If calculation is not specified a default is set.
        """
        selectors = action.get("selectors", {})
        if "codename" in action:
            selectors.update({action["property"]: action["codename"]})
        calculation = "Avg" if "calculation" not in action else action["calculation"]
        return selectors, calculation

    @staticmethod
    def _get_statistical_value(
        dframe: pd.DataFrame,
        prop: str,
        calculation: str,
        selectors: dict = None,
    ) -> float:
        """
        Retrive statistical value from the property statistic dataframe
        """

        dframe = dframe[dframe["PROPERTY"] == prop].copy()

        if selectors is not None:
            for selector, value in selectors.items():
                dframe = dframe[dframe[selector] == value]

        if len(dframe.index) > 1:
            print(dframe)
            raise Exception("Ambiguous result, multiple rows meet conditions")

        return dframe.iloc[0][calculation]



"""Module for generating design matrices that can be run by DESIGN2PARAMS
and DESIGN_KW in FMU/ERT.
"""
from collections import OrderedDict
from pathlib import Path

import pandas as pd
import numpy

from fmu.tools.sensitivities import design_distributions as design_dist


class DesignMatrix:
    """Class for design matrix in FMU. Can contain a onebyone design
    or a full montecarlo design.

    Attributes:
        designvalues (pd.DataFrame): design matrix on standard fmu format
            contains columns 'REAL' (realization number), and if a onebyone
            design, also columns 'SENSNAME' and 'SENSCASE'
        defaultvalues (OrderedDict): default values for design
        backgroundvalues (pd.DataFrame): Used when background parameters are
            not constant. Either a set is sampled from specified distributions
            or they are read from a file.
    """

    def __init__(self):
        """
        Placeholders for:
        designvalues: dataframe with parameters that varies
        defaultvalues: dictionary of default/base case values
        backgroundvalues: dataframe with background parameters
        seedvalues: list of seed values
        """
        self.designvalues = pd.DataFrame(columns=["REAL"])
        self.defaultvalues = OrderedDict()
        self.backgroundvalues = None
        self.seedvalues = None

    def reset(self):
        """Resets DesignMatrix to empty. Necessary iin case method generate
        is used several times for same instance of DesignMatrix"""
        self.designvalues = pd.DataFrame(columns=["REAL"])
        self.defaultvalues = OrderedDict()
        self.backgroundvalues = None
        self.seedvalues = None

    def generate(self, inputdict):
        """Generating design matrix from input dictionary in specific
        format. Adding default values and background values if existing.
        Looping through sensitivities and adding them to designvalues.

        Args:
            inputdict (OrderedDict): input parameters for design
        """
        self.reset()  # Emptying if regenerating matrix

        if "distribution_seed" in inputdict.keys():
            if inputdict["distribution_seed"]:
                numpy.random.seed(inputdict["distribution_seed"])

        # Reading default values
        default_dict = inputdict["defaultvalues"]
        self.set_defaultvalues(default_dict)

        max_reals = _find_max_realisations(inputdict)

        # Reading or generating rms seed values
        if "seeds" in inputdict.keys():
            self.add_seeds(inputdict["seeds"], max_reals)

        # If background values used - read or generate
        if "background" in inputdict.keys():
            self.add_background(inputdict["background"], max_reals)

        if inputdict["designtype"] == "onebyone":
            self.designvalues["SENSNAME"] = None
            self.designvalues["SENSCASE"] = None
            counter = 0
            for key in inputdict["sensitivities"].keys():
                sens = inputdict["sensitivities"][key]
                if "numreal" in sens.keys():
                    numreal = sens["numreal"]
                else:
                    numreal = inputdict["repeats"]
                if sens["senstype"] == "ref":
                    sensitivity = SingleRealisationReference(key)
                    sensitivity.generate(range(counter, counter + numreal))
                    counter += numreal
                    self._add_sensitivity(sensitivity)
                elif sens["senstype"] == "background":
                    sensitivity = BackgroundSensitivity(key)
                    sensitivity.generate(range(counter, counter + numreal))
                    counter += numreal
                    self._add_sensitivity(sensitivity)
                elif sens["senstype"] == "seed":
                    if self.seedvalues is None:
                        raise ValueError(
                            "No seed values available to use for seed sensitivity"
                        )
                    sensitivity = SeedSensitivity(key)
                    sensitivity.generate(
                        range(counter, counter + numreal),
                        sens["seedname"],
                        self.seedvalues,
                        sens["parameters"],
                    )
                    counter += numreal
                    self._add_sensitivity(sensitivity)
                elif sens["senstype"] == "scenario":
                    sensitivity = ScenarioSensitivity(key)
                    for casekey in sens["cases"].keys():
                        case = sens["cases"][casekey]
                        temp_case = ScenarioSensitivityCase(casekey)
                        temp_case.generate(
                            range(counter, counter + numreal), case, self.seedvalues
                        )
                        sensitivity.add_case(temp_case)
                        counter += numreal
                    self._add_sensitivity(sensitivity)
                elif sens["senstype"] == "dist":
                    sensitivity = MonteCarloSensitivity(key)
                    sensitivity.generate(
                        range(counter, counter + numreal),
                        sens["parameters"],
                        self.seedvalues,
                        sens["correlations"],
                    )
                    counter += numreal
                    self._add_sensitivity(sensitivity)
                elif sens["senstype"] == "extern":
                    sensitivity = ExternSensitivity(key)
                    sensitivity.generate(
                        range(counter, counter + numreal),
                        sens["extern_file"],
                        sens["parameters"],
                        self.seedvalues,
                    )
                    counter += numreal
                    self._add_sensitivity(sensitivity)
                print("Added sensitivity :", sensitivity.sensname)
            if "background" in inputdict.keys():
                self._fill_with_background_values()
            self._fill_with_defaultvalues()

            if "dependencies" in inputdict.keys():
                self._fill_derived_params(inputdict["dependencies"])

            if "decimals" in inputdict.keys():
                self._set_decimals(inputdict["decimals"])
            # Re-order columns
            start_cols = ["REAL", "SENSNAME", "SENSCASE", "RMS_SEED"]
            self.designvalues = self.designvalues[
                [col for col in start_cols if col in self.designvalues]
                + [col for col in self.designvalues if col not in start_cols]
            ]
        else:
            raise ValueError(
                "Generation of DesignMatrix only"
                "implemented for type onebyone"
                "In general_input designtype is"
                "set to {}".format(inputdict["designtype"])
            )

    def to_xlsx(
        self, filename, designsheet="DesignSheet01", defaultsheet="DefaultValues"
    ):
        """Writing design matrix to excel workfbook on standard fmu format
        to be used in FMU/ERT by DESIGN2PARAMS and DESIGN_KW

        Args:
            filename (str): output filename (extension .xlsx)
            designsheet (str): name of excel sheet containing design matrix
                (optional, defaults to 'DesignSheet01')
            defaultsheet (str): name of excel sheet containing default
                values (optional, defaults to 'DefaultValues')
        """
        # pylint: disable=abstract-class-instantiated
        basename = Path(filename).stem
        extension = Path(filename).suffix
        if extension != ".xlsx":
            filename = basename + ".xlsx"
            print(
                "Warning: Output filename did not have extension .xlsx "
                "but the export format is Excel .xlsx .  "
                "Changing outputname to {}".format(filename)
            )

        xlsxwriter = pd.ExcelWriter(filename, engine="openpyxl")
        self.designvalues.to_excel(
            xlsxwriter, sheet_name=designsheet, index=False, header=True
        )
        # Default values from OrderdDictionay to pandas dataframe
        defaults = pd.DataFrame(columns=["defaultparameters", "defaultvalue"])
        defaults["defaultparameters"] = self.defaultvalues.keys()
        defaults["defaultvalue"] = self.defaultvalues.values()
        defaults.to_excel(
            xlsxwriter, sheet_name=defaultsheet, index=False, header=False
        )
        xlsxwriter.save()
        print(
            "A total of {} realizations were generated".format(
                len(self.designvalues["REAL"])
            )
        )
        print("Designmatrix written to {}".format(filename))

    def set_defaultvalues(self, defaults):
        """Add default values

        Args:
            defaults (OrderedDict): (key, value) is (parameter_name, value)
        """
        self.defaultvalues = defaults

    def add_seeds(self, seeds, max_reals):
        """Adding background as specified in dictionary.
        Either from external file or from distributions in background
        dictionary

        Args:
            back_dict (OrderedDict): how to generate background values
            max_values (int): number of background values to generate
        """
        if seeds in [None, "None"]:
            self.seedvalues = None
            print("seeds is set to None in general_input")
        elif seeds.lower() == "default":
            self.seedvalues = [item + 1000 for item in range(max_reals)]
        elif Path(seeds).is_file():
            self.seedvalues = _seeds_from_extern(seeds, max_reals)
        else:
            raise ValueError(
                "Valid choices for seeds are None, "
                '"default" or an existing filename. '
                "Neither was found in this case. seeds "
                "had been specified as {} .".format(seeds)
            )

    def add_background(self, back_dict, max_values):
        """Adding background as specified in dictionary.
        Either from external file or from distributions in background
        dictionary

        Args:
            back_dict (OrderedDict): how to generate background values
            max_values (int): number of background values to generate
        """
        if back_dict is None:
            self.backgroundvalues = None
        elif "extern" in back_dict.keys():
            self.backgroundvalues = _parameters_from_extern(back_dict["extern"])
        elif "parameters" in back_dict.keys():
            self._add_dist_background(back_dict, max_values)

    def background_to_excel(self, filename, backgroundsheet="Background"):
        """Writing background values to an Excel spreadsheet

        Args:
            filename (str): output filename (extension .xlsx)
            backgroundsheet (str): name of excel sheet
        """
        # pylint: disable=abstract-class-instantiated
        xlsxwriter = pd.ExcelWriter(filename, engine="openpyxl")
        self.backgroundvalues.to_excel(
            xlsxwriter, sheet_name=backgroundsheet, index=False, header=True
        )
        xlsxwriter.save()
        print("Backgroundvalues written to {}".format(filename))

    def _add_sensitivity(self, sensitivity):
        """Adding a sensitivity to the design

        Args:
            sensitivity of class Scenario, MonteCarlo or Extern
        """
        existing_values = self.designvalues.copy()
        self.designvalues = existing_values.append(sensitivity.sensvalues, sort=False)

    def _fill_with_background_values(self):
        """Substituting NaNs with background values if existing.
        background values not in design are added as separate colums
        """
        if self.backgroundvalues is not None:
            grouped = self.designvalues.groupby(["SENSNAME", "SENSCASE"], sort=False)
            result_values = pd.DataFrame()
            for sensname, case in grouped:
                temp_df = case.reset_index()
                temp_df.fillna(self.backgroundvalues, inplace=True)
                temp_df.set_index("index")
                for key in self.backgroundvalues.keys():
                    if key not in case.keys():
                        temp_df[key] = self.backgroundvalues[key]
                        if len(temp_df) > len(self.backgroundvalues):
                            raise ValueError(
                                "Provided number of background values "
                                "{} is smaller than number"
                                " of realisations for sensitivity {}".format(
                                    len(self.backgroundvalues), sensname
                                )
                            )
                    else:
                        if len(temp_df) > len(self.backgroundvalues):
                            print(
                                "Provided number of background values "
                                "({}) is smaller than number"
                                " of realisations for sensitivity {}"
                                " and parameter {}. "
                                "Will be filled with default values.".format(
                                    len(self.backgroundvalues), sensname, key
                                )
                            )
                existing_values = result_values.copy()
                result_values = existing_values.append(temp_df, sort=False)

            result_values = result_values.drop(["index"], axis=1)
            self.designvalues = result_values

    def _fill_with_defaultvalues(self):
        """Filling NaNs with default values"""
        for key in self.designvalues.keys():
            if key in self.defaultvalues.keys():
                self.designvalues[key].fillna(self.defaultvalues[key], inplace=True)
            elif key not in ["REAL", "SENSNAME", "SENSCASE", "RMS_SEED"]:
                raise LookupError(
                    "No defaultvalues given for parameter {} " "".format(key)
                )

    def _fill_derived_params(self, depend_dict):
        for from_param in depend_dict.keys():
            if from_param in self.designvalues.keys():
                for param in depend_dict[from_param]["to_params"].keys():
                    self.designvalues[param] = numpy.nan
                    for index in range(len(depend_dict[from_param]["from_values"])):
                        fill_these = depend_dict[from_param]["from_values"][index]
                        fill_data = depend_dict[from_param]["to_params"][param][index]
                        self.designvalues[param].mask(
                            self.designvalues[from_param] == fill_these,
                            fill_data,
                            inplace=True,
                        )
                    if self.designvalues[param].isnull().any():
                        raise ValueError(
                            "Column for derived parameter {} "
                            "contains NaN. Check input "
                            "defining dependencies. "
                            "Could be Wrong values or that "
                            "values for input variable  in "
                            "dependencies sheet "
                            "should be specified as strings.".format(param)
                        )

    def _add_dist_background(self, back_dict, numreal):
        """Drawing background values from distributions
        specified in dictionary

        Args:
            back_dict (OrderedDict): parameters and distributions
            numreal (int): Number of samples to generate
        """
        assert isinstance(
            numreal, int
        ), f"numreal must be an integer, got {type(numreal)} with value {numreal}"
        mc_background = MonteCarloSensitivity("background")
        mc_background.generate(
            range(numreal), back_dict["parameters"], "None", back_dict["correlations"]
        )
        mc_backgroundvalues = mc_background.sensvalues

        # Rounding of background values as specified
        if "decimals" in back_dict.keys():
            for key in back_dict["decimals"].keys():
                if design_dist.is_number(mc_backgroundvalues[key].iloc[0]):
                    mc_backgroundvalues[key] = (
                        mc_backgroundvalues[key]
                        .astype(float)
                        .round(int(back_dict["decimals"][key]))
                    )
                else:
                    raise ValueError("Cannot round a string parameter")
        self.backgroundvalues = mc_backgroundvalues.copy()

    def _set_decimals(self, dict_decimals):
        """Rounding to specified number of decimals

        Args:
            dict_decimals (dictionary): (key, value)s are (param, decimals)
        """
        for key in self.designvalues.keys():
            if key in dict_decimals.keys():
                if design_dist.is_number(self.designvalues[key].iloc[0]):
                    self.designvalues[key] = (
                        self.designvalues[key]
                        .astype(float)
                        .round(int(dict_decimals[key]))
                    )
                else:
                    raise ValueError("Cannot round a string parameter {}".format(key))


class SeedSensitivity:
    """
    A seed sensitivity is normally the reference for one by one sensitivities
    It contains a list of seeds to be repeated for each sensitivity
    The parameter name is hardcoded to RMS_SEED
    It will be assigned the sensname 'p10_p90' which will be written to
    the SENSCASE column in the output.

    Attributes:
        sensname (str): name of sensitivity
        sensvalues (pd.DataFrame):  design values for the sensitivity

    """

    # pylint: disable=too-few-public-methods
    def __init__(self, sensname):
        """Args:
        sensname (str): Name of sensitivity.
            Defines SENSNAME in design matrix
        """
        self.sensname = sensname
        self.sensvalues = None

    def generate(self, realnums, seedname, seedvalues, parameters):
        """Generates parameter values for a seed sensitivity

        Args:
            realnums (list): list of integers with realization numbers
            seedname (str): name of seed parameter to add
            seedvalues (list): list of integer seedvalues
            parameters (OrderedDict): parameter names and
                distributions or values.
        """
        assert isinstance(
            seedvalues, list
        ), f"seedvalues must be a list, got {seedvalues}"

        self.sensvalues = pd.DataFrame(index=realnums)
        self.sensvalues[seedname] = seedvalues[0 : len(realnums)]

        if parameters is not None:
            for key in parameters.keys():
                dist_name = parameters[key][0].lower()
                constant = parameters[key][1]
                if dist_name != "const":
                    raise ValueError(
                        'A sensitivity of type "seed" can only have '
                        "additional parameters where dist_name is "
                        '"const". Check sensitivity {}"'.format(self.sensname)
                    )
                self.sensvalues[key] = constant

        self.sensvalues["REAL"] = realnums
        self.sensvalues["SENSNAME"] = self.sensname
        self.sensvalues["SENSCASE"] = "p10_p90"


class SingleRealisationReference:
    """
    The class is used in set-ups where one wants a single realisation
    containing only default values as a reference, but the realisation
    itself is not included in a sensitivity.
    Typically used when RMS_SEED is not a parameter.
    SENSCASE will be set to 'ref' in design matrix, to flag that it should be
    excluded as a sensitivity in the plot.

    Attributes:
        sensname (str): name of sensitivity
        sensvalues (pd.DataFrame):  design values for the sensitivity

    """

    # pylint: disable=too-few-public-methods
    def __init__(self, sensname):
        """Args:
        sensname (str): Name of sensitivity.
            Defines SENSNAME in design matrix
        """
        self.sensname = sensname
        self.sensvalues = None

    def generate(self, realnums):
        """Generates realisation number only

        Args:
            realnums (list): list of intergers with realization numbers
        """
        self.sensvalues = pd.DataFrame(index=realnums)
        self.sensvalues["REAL"] = realnums
        self.sensvalues["SENSNAME"] = self.sensname
        self.sensvalues["SENSCASE"] = "ref"


class BackgroundSensitivity:
    """
    The class is used in set-ups where one sensitivities
    are run on top of varying background parameters.
    Typically used when RMS_SEED is not a parameter, so the reference
    for tornadoplots will be the realisations with all parameters
    at their default values except the background parameters.
    SENSCASE will be set to 'p10_p90' in design matrix.

    Attributes:
        sensname (str): name of sensitivity
        sensvalues (pd.DataFrame):  design values for the sensitivity

    """

    # pylint: disable=too-few-public-methods
    def __init__(self, sensname):
        """Args:
        sensname (str): Name of sensitivity.
            Defines SENSNAME in design matrix
        """
        self.sensname = sensname
        self.sensvalues = None

    def generate(self, realnums):
        """Generates realisation number only

        Args:
            realnums (list): list of intergers with realization numbers
        """
        self.sensvalues = pd.DataFrame(index=realnums)
        self.sensvalues["REAL"] = realnums
        self.sensvalues["SENSNAME"] = self.sensname
        self.sensvalues["SENSCASE"] = "p10_p90"


class ScenarioSensitivity:
    """Each design can contain one or several single sensitivities of type
    Seed, MonteCarlo or Scenario.
    Each ScenarioSensitivity can contain 1-2 ScenarioSensitivityCases.

    The ScenarioSensitivity class is used for sensitivities where all
    realizatons in a ScenarioSensitivityCase have identical values
    but one or more parameter has a different values from the other
    ScenarioSensitivityCase.

    Exception is the seed value and the special case where
    varying background parameters are specified. Then these are varying
    within the case.

    Attributes:
        case1 (ScenarioSensitivityCase): first case, e.g. 'low case'
        case2 (ScenarioSensitivityCase): second case, e.g. 'high case'
        sensvalues (pd.DataFrame): design values for the sensitivity, containing
           1-2 cases
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, sensname):
        """
        Args:
            sensname (str): Name of sensitivity.
                Equals SENSNAME in design matrix
        """
        self.sensname = sensname
        self.case1 = None
        self.case2 = None
        self.sensvalues = None

    def add_case(self, senscase):
        """
        Adds a ScenarioSensitivityCase instance
        to a ScenarioSensitivity object.

        Args:
            senscase (ScenarioSensitivityCase):
                Equals SENSCASE in design matrix.
        """
        if self.case1 is not None:  # Case 1 has been read, this is case2
            if (
                "REAL" in senscase.casevalues.keys()
                and "SENSCASE" in senscase.casevalues.keys()
            ):
                self.case2 = senscase
                senscase.casevalues["SENSNAME"] = self.sensname
                self.sensvalues = pd.concat(
                    [self.sensvalues, senscase.casevalues], sort=True
                )
        else:  # This is the first case
            if (
                "REAL" in senscase.casevalues.keys()
                and "SENSCASE" in senscase.casevalues.keys()
            ):
                self.case1 = senscase
                self.sensvalues = senscase.casevalues.copy()
                self.sensvalues["SENSNAME"] = self.sensname


class ScenarioSensitivityCase:
    """Each ScenarioSensitivity can contain one or
    two ScenarioSensitivityCases.

    The 1-2 cases are typically 'low' and 'high' cases for one or
    a set of  parameters, where all realisatons in
    the case have identical values except the seed value
    and in special cases specified background values which may
    vary within the case.

    One or two ScenarioSensitivityCase instances can be added to each
    ScenarioSensitivity object.

    Attributes:
        casename (str): name of the sensitivity case,
            equals SENSCASE in design matrix.
        casevalues (pd.DataFrame): parameters and values
            for the sensitivity with realisation numbers as index.

    """

    # pylint: disable=too-few-public-methods
    def __init__(self, casename):
        self.casename = casename
        self.casevalues = None

    def generate(self, realnums, parameters, seedvalues):
        """Generate casevalues for the ScenarioSensitivityCase

        Args:
            realnums (list): list of realizaton numbers for the case
            parameters (OrderedDict):
                dictionary with parameter names and values
            seeds (str): default or None
        """

        self.casevalues = pd.DataFrame(columns=parameters.keys(), index=realnums)
        for key in parameters.keys():
            self.casevalues[key] = parameters[key]
        self.casevalues["REAL"] = realnums
        self.casevalues["SENSCASE"] = self.casename

        if seedvalues:
            self.casevalues["RMS_SEED"] = seedvalues[: len(realnums)]


class MonteCarloSensitivity:
    """
    For a MonteCarloSensitivity one or several parameters
    are drawn from specified distributions with or without correlations.
    A MonteCarloSensitivity can only contain
    one case, where the name SENSCASE is automatically set to 'p10_p90' in the
    design matrix to flag that p10_p90 should be calculated in TornadoPlot.

    Attributes:
        sensname (str):  name for the sensitivity.
            Equals SENSNAME in design matrix.
        sensvalues (pd.DataFrame):  parameters and values for the sensitivity
            with realisation numbers as index.
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, sensname):
        self.sensname = sensname
        self.sensvalues = None

    def generate(self, realnums, parameters, seedvalues, corrdict):
        """Generates parameter values by drawing from
        defined distributions.

        Args:
            realnums (list): list of integers with realization numbers
            parameters (OrderedDict):
                dictionary of parameters and distributions
            seeds (str): default or None
            corrdict(OrderedDict): correlation info
        """
        self.sensvalues = pd.DataFrame(columns=parameters.keys(), index=realnums)
        numreals = len(realnums)
        if corrdict is None:
            for key in parameters.keys():
                dist_name = parameters[key][0].lower()
                dist_params = parameters[key][1]
                try:
                    self.sensvalues[key] = design_dist.draw_values(
                        dist_name, dist_params, numreals
                    )
                except ValueError as error:
                    raise ValueError(
                        "Problem with parameter {} in sensitivity "
                        "with sensname {}: {}.".format(
                            key, self.sensname, error.args[0]
                        )
                    ) from error
        else:  # Some or all parameters are correlated
            df_params = pd.DataFrame.from_dict(
                parameters,
                orient="index",
                columns=["dist_name", "dist_params", "corr_sheet"],
            )
            df_params["corr_sheet"].fillna("nocorr", inplace=True)
            df_params.reset_index(inplace=True)
            df_params.rename(columns={"index": "param_name"}, inplace=True)
            param_groups = df_params.groupby(["corr_sheet"])

            for correl, group in param_groups:
                param_dict = OrderedDict()
                for _, row in group.iterrows():
                    param_dict[row["param_name"]] = [
                        row["dist_name"],
                        row["dist_params"],
                    ]
                if not correl == "nocorr":
                    if len(group) == 1:
                        _printwarning(correl)
                    df_correlations = design_dist.read_correlations(corrdict, correl)
                    multivariate_parameters = df_correlations.index.values
                    cov_matrix = design_dist.make_covariance_matrix(df_correlations)
                    normalscoremeans = len(multivariate_parameters) * [0]
                    normalscoresamples = numpy.random.multivariate_normal(
                        normalscoremeans, cov_matrix, size=numreals
                    )
                    normalscoresamples_df = pd.DataFrame(
                        data=normalscoresamples, columns=multivariate_parameters
                    )
                    for key in param_dict.keys():
                        dist_name = param_dict[key][0].lower()
                        dist_parameters = param_dict[key][1]
                        if key in multivariate_parameters:
                            try:
                                self.sensvalues[key] = design_dist.draw_values(
                                    dist_name,
                                    dist_parameters,
                                    numreals,
                                    normalscoresamples_df[key],
                                )
                            except ValueError as error:
                                raise ValueError(
                                    "Problem in sensitivity "
                                    "with sensname {} for "
                                    "parameter {}: {}.".format(
                                        self.sensname, key, error.args[0]
                                    )
                                ) from error
                        else:
                            raise ValueError(
                                "Parameter{} specified with correlation "
                                "matrix {} but is not listed in "
                                "that sheet".format(key, correl)
                            )
                else:  # group nocorr where correlation matrix is not defined
                    for key in param_dict.keys():
                        dist_name = param_dict[key][0].lower()
                        dist_parameters = param_dict[key][1]
                        try:
                            self.sensvalues[key] = design_dist.draw_values(
                                dist_name, dist_parameters, numreals
                            )
                        except ValueError as error:
                            raise ValueError(
                                "Problem in sensitivity "
                                "with sensname {} for parameter "
                                "{}: {}.".format(self.sensname, key, error.args[0])
                            ) from error

        if self.sensname != "background":
            self.sensvalues["REAL"] = realnums
            self.sensvalues["SENSNAME"] = self.sensname
            self.sensvalues["SENSCASE"] = "p10_p90"
            if "RMS_SEED" not in self.sensvalues.keys() and seedvalues:
                self.sensvalues["RMS_SEED"] = seedvalues[: len(realnums)]


class ExternSensitivity:
    """
    Used when reading parameter values from a file
    Assumed to be used with monte carlo type sensitivities and
    will hence write 'p10_p90' as SENSCASE in output designmatrix

    Attributes:
        sensname (str): Name of sensitivity.
            Defines SENSNAME in design matrix
        sensvalues (pd.DataFrame):  design values for the sensitivity

    """

    # pylint: disable=too-few-public-methods
    def __init__(self, sensname):
        self.sensname = sensname
        self.sensvalues = None

    def generate(self, realnums, filename, parameters, seedvalues):
        """Reads parameter values for a monte carlo sensitivity
        from file

        Args:
            realnums (list): list of intergers with realization numbers
            filename (str): path where to read values from
            parameters (list): list with parameter names
            seeds (str): default or None
        """
        self.sensvalues = pd.DataFrame(columns=parameters, index=realnums)
        extern_values = _parameters_from_extern(filename)
        if len(realnums) > len(extern_values):
            raise ValueError(
                "Number of realisations {} specified for "
                "sensitivity {} is larger than rows in "
                "file {}".format(len(realnums), self.sensname, filename)
            )
        for param in parameters:
            if param in extern_values.keys():
                self.sensvalues[param] = list(extern_values[param][: len(realnums)])
            else:
                raise ValueError("Parameter {} not in external file".format(param))
        self.sensvalues["REAL"] = realnums
        self.sensvalues["SENSNAME"] = self.sensname
        self.sensvalues["SENSCASE"] = "p10_p90"

        if seedvalues:
            self.sensvalues["RMS_SEED"] = seedvalues[: len(realnums)]


# Support functions used with several classes


def _parameters_from_extern(filename):
    """Read parameter values or background values
    from specified file. Format either Excel ('xlsx')
    or csv.

    Args:
        filename (str): path to file
    """
    if str(filename).endswith(".xlsx"):
        parameters = pd.read_excel(filename, engine="openpyxl")
        parameters.dropna(axis=0, how="all", inplace=True)
        parameters = parameters.loc[:, ~parameters.columns.str.contains("^Unnamed")]
    elif str(filename).endswith(".csv"):
        parameters = pd.read_csv(filename)
    else:
        raise ValueError(
            "External file with parameter values should "
            "be on Excel or csv format "
            "and end with .xlsx or .csv"
        )
    return parameters


def _seeds_from_extern(filename, max_reals):
    """Read parameter values or background values
    from specified file. Format either Excel ('xlsx')
    or csv.

    Args:
        filename (str): path to file
    """
    if str(filename).endswith(".xlsx"):
        df_seeds = pd.read_excel(filename, header=None, engine="openpyxl")
        df_seeds.dropna(axis=0, how="all", inplace=True)
        df_seeds.dropna(axis=1, how="all", inplace=True)
        seed_numbers = df_seeds[df_seeds.columns[0]].tolist()
    elif str(filename).endswith(".csv") or str(filename).endswith(".txt"):
        df_seeds = pd.read_csv(filename, header=None)
        seed_numbers = df_seeds[df_seeds.columns[0]].tolist()
    else:
        raise ValueError(
            "External file with seed values should "
            "be on Excel or csv format "
            "and end with .xlsx .csv or .txt"
        )

    if len(seed_numbers) < max_reals:
        print(
            "Provided number of seed values in external file {} "
            "is lower than the maximum number of realisations "
            "found for the design {}, and is for those "
            "sensitivities used repeatedly. "
        )
        seed_numbers = [
            seed_numbers[item % len(seed_numbers)] for item in range(max_reals)
        ]
    return seed_numbers


def _find_max_realisations(inputdict):
    """Finds the maximum number of realisations
    in a sensitivity case"""
    max_reals = inputdict["repeats"]
    for key in inputdict["sensitivities"].keys():
        sens = inputdict["sensitivities"][key]
        if "numreal" in sens.keys():
            max_reals = max(sens["numreal"], max_reals)
    return max_reals


def _printwarning(corrgroup):
    print(
        "#######################################################\n"
        "fmudesign Warning:                                     \n"
        "Using designinput sheets where "
        "corr_sheet is only specified for one parameter "
        "will cause non-correlated parameters .\n"
        "ONLY ONE PARAMETER WAS SPECIFIED TO USE CORR_SHEET {}\n"
        "\n"
        "Note change in how correlated parameters are specified \n"
        "from fmudeisgn version 1.0.1 in August 2019 :\n"
        "Name of correlation sheet must be specified for each "
        "parameter in correlation matrix. \n"
        "This to enable use of several correlation sheets. "
        "This also means non-correlated parameters do not "
        "have to be included in correlation matrix. \n "
        "See documentation: \n"
        "https://sdp.equinor.com/wikidocs/FMU/lib/fmu/tools/"
        "html/examples.html#create-design-matrix-for-"
        "one-by-one-sensitivities\n"
        "\n"
        "####################################################\n"
        "".format(corrgroup)
    )



"""
This private module in qcforward is used to parse data on a general level.

The resulting data will be stored as class instance attributes, e.g. self._grid

"""
import re
from glob import glob
from os.path import join

import xtgeo
from fmu.tools._common import _QCCommon

CMN = _QCCommon()


class QCData(object):
    """
    This is a class which parse/reads and stores some common data
    like 3D grids, maps etc.

    It shall be semi-agnostic to where data comes from, i.e. accept both files
    and RMS input.

    By having it as a class one can store different datasets and compare them later
    if needed. E.g. if one need to compare two grid models, they will be members
    of two different instances of the QCData()
    """

    def __init__(self):
        self._reuse = []
        self._xtgdata = {}  # All XTGeo data, to be reused
        self._path = "."  # usually not needed to set explisit
        self._project = None  # pointing to  RMS project ID if any
        self._grid = None  # A XTGeo Grid()
        self._gridprops = None  # XTGeo GridProperties() for multiple props
        self._wells = None  # XTGeo Wells() instance (multiple wells)
        self._bwells = None
        self._surfaces = None  # XTGeo Surfaces() instance (in prep)

        self._set_xtgdata_keys()

    # Properties
    # ==================================================================================

    @property
    def project(self):
        """Project attribute as Roxar project instance"""
        return self._project

    @property
    def grid(self):
        """Grid attribute as XTGeo Grid() instance"""
        return self._grid

    @property
    def gridprops(self):
        """GridProperties attribute as XTGeo GridProperties() instance"""
        return self._gridprops

    @property
    def wells(self):
        """Wells attribute as XTGeo Wells() instance"""
        return self._wells

    @property
    def bwells(self):
        """Wells attribute as XTGeo BlockedWells() instance"""
        return self._bwells

    @property
    def xtgdata(self):
        """All xtgeo based data as a dictionay"""
        return self._xtgdata

    # Class methods:
    # ==================================================================================

    def _set_xtgdata_keys(self):
        self._xtgdata["grid"] = {}
        self._xtgdata["gridprops"] = {}
        self._xtgdata["wells"] = {}
        self._xtgdata["bwells"] = {}

    def parse(
        self,
        data=None,
        project=None,
        reuse=False,
        wells_settings=None,
    ):
        """Parse the actual data, such as grids, gridprops etc.

        Args:
            data (dict): The input data dictionary
                fine grained control, e.g. ["grid", "gridprops", "wells"]
            project (obj or str): For usage inside RMS
            reuse (bool or list): If True, do not reread grid and gridprops. Can also be
                a list referring to key names.
            wells_settings (dict): More granular settings for wells which may speed
                up reading and/or execution. If set, this is a dictionary with keys
                ``lognames``, ``depthrange``, ``resample``

        """

        CMN.verbosity = data.get("verbosity")

        if isinstance(reuse, bool):
            reuse = ["grid", "gridprops", "wells"] if reuse else []

        self.set_path(data.get("path"))
        self.parse_project(project)

        if "grid" in data:
            self.read_grid(data["grid"], reuse)

        if "gridprops" in data:
            self.read_gridprops(data["gridprops"], data.get("grid"), reuse)

        for welltype in ["wells", "bwells"]:
            if welltype in data:
                self.read_wells(data[welltype], welltype, wells_settings, reuse)

    def set_path(self, path):
        """General path prefix settings"""
        self._path = path

    def parse_project(self, project=None):
        """Get the RoxarAPI project magics"""

        if project is not None:
            rox = xtgeo.RoxUtils(project)
            self._project = rox.project
        else:
            self._project = None

    def read_grid(self, gridname, reuse):
        """Read 3D grid (which is required), from file or RMS"""

        gridname = gridname if self._project is not None else join(self._path, gridname)
        CMN.print_debug(f"GRIDNAME: {gridname}")

        CMN.print_info("Reading grid geometry...")
        if ("grid" not in reuse) or (gridname not in self._xtgdata["grid"]):
            self._grid = (
                xtgeo.Grid(gridname)
                if self._project is None
                else xtgeo.grid_from_roxar(self._project, gridname)
            )
            self._xtgdata["grid"][gridname] = self._grid
            self._xtgdata["gridprops"][gridname] = {}
        else:
            CMN.print_info(f"Reusing grid {gridname}")
            self._grid = self._xtgdata["grid"][gridname]

    def read_gridprops(self, gridprops, gridname=None, reuse=None):
        """Read 3D grid props, from file or RMS."""

        gridname = gridname if self._project is not None else join(self._path, gridname)

        CMN.print_info("Reading grid properties...")
        gprops = []
        if "gridprops" in reuse:
            reused_gprops, gridprops = self._reuse_gridprops(gridprops, gridname)
            gprops = reused_gprops

        if self._project is None:
            for gprop in gridprops:
                if isinstance(gprop, list):
                    pname, pfile = gprop
                else:
                    pfile = gprop
                    pname = "unknown"

                gridproppath = join(self._path, pfile)

                xtg_gprop = xtgeo.gridproperty_from_file(
                    gridproppath, name=pname, grid=self.grid
                )
                xtg_gprop.name = pname if pname != "unknown" else pfile
                gprops.append(xtg_gprop)
                if isinstance(gprop, list):
                    self._xtgdata["gridprops"][gridname][tuple(gprop)] = xtg_gprop
                else:
                    self._xtgdata["gridprops"][gridname][gprop] = xtg_gprop
        else:
            # read from RMS/ROXAPI
            for pname in gridprops:
                xtg_gprop = xtgeo.gridproperty_from_roxar(
                    self._project, gridname, pname
                )
                gprops.append(xtg_gprop)
                self._xtgdata["gridprops"][gridname][pname] = xtg_gprop

        self._gridprops = xtgeo.GridProperties()
        self._gridprops.append_props(gprops)

    def _well_preparations(self, wells):
        """Account for wildcards and regex in wellist"""

        input_wells = []
        if self._project is None:
            for welldata in wells:
                # fields may contain wildcards for "globbing"
                for wellentry in glob(join(self._path, welldata)):
                    input_wells.append(wellentry)
        else:
            # roxar API input:
            wnames = wells.get("names", [".*$"])
            rmswells = [wll.name for wll in self._project.wells]

            CMN.print_debug(f"All RMS wells: {rmswells}")
            CMN.print_debug(f"Data wells to match: {wnames}")

            for rmswell in rmswells:
                if any(re.match(wreg + "$", rmswell) for wreg in wnames):
                    input_wells.append(rmswell)

        return input_wells

    def read_wells(self, wells, welltype="wells", settings=False, reuse=None):
        """Reading wells"""

        settings = settings if settings else {}
        wellist = self._well_preparations(wells)

        CMN.print_info("Reading wells...")
        xtg_wells = []
        if "wells" in reuse:
            reused_wells, wellist = self._reuse_wells(wellist, welltype)
            xtg_wells = reused_wells

        for well in wellist:
            try:
                if welltype == "wells":
                    mywell = (
                        xtgeo.well_from_file(
                            well, lognames=settings.get("lognames", "all")
                        )
                        if self._project is None
                        else xtgeo.well_from_roxar(
                            project=self._project,
                            name=well,
                            lognames=settings.get("lognames", "all"),
                            logrun=wells.get("logrun", "log"),
                            trajectory=wells.get("trajectory", "Drilled trajectory"),
                        )
                    )
                else:
                    mywell = (
                        xtgeo.blockedwell_from_file(well)
                        if self._project is None
                        else xtgeo.blockedwell_from_roxar(
                            project=self._project,
                            gname=wells.get("grid", "Geogrid"),
                            bwname=wells.get("bwname", "BW"),
                            wname=well,
                            lognames=settings.get("lognames", "all"),
                        )
                    )
                xtg_wells.append(mywell)
                self._xtgdata[welltype][well] = mywell

            except ValueError as verr:
                print(f"Could not read well {well}: {verr}")

        CMN.print_debug(f"All valid welldata: {xtg_wells}")

        for mywell in xtg_wells:
            if "depthrange" in settings and settings["depthrange"] is not None:
                tmin, tmax = settings["depthrange"]
                mywell.limit_tvd(tmin, tmax)
            if "rescale" in settings and settings["rescale"] is not None:
                mywell.rescale(settings["rescale"])

        if xtg_wells:
            if welltype == "wells":
                self._wells = xtgeo.Wells()
                self._wells.wells = xtg_wells
            else:
                self._bwells = xtgeo.BlockedWells()
                self._bwells.wells = xtg_wells
        else:
            raise RuntimeError("No wells read, wrong settings?")

    def _reuse_gridprops(self, gridprops, gridname):
        """
        Identify which gridprops are available for reusing, reusable
        gridprops and new gridprops are returned in separate lists
        """
        new_gprops = []
        reused_gprops = []

        for elem in gridprops:
            if isinstance(elem, list):
                if tuple(elem) not in self._xtgdata["gridprops"][gridname]:
                    new_gprops.append(elem)
                reused_gprops = [
                    value
                    for key, value in self._xtgdata["gridprops"][gridname].items()
                    if list(key) in gridprops
                ]
            else:
                if elem not in self._xtgdata["gridprops"][gridname]:
                    new_gprops.append(elem)
                reused_gprops = [
                    value
                    for key, value in self._xtgdata["gridprops"][gridname].items()
                    if key in gridprops
                ]

        CMN.print_info(f"Reusing gridprops: {[x.name for x in reused_gprops]}")
        CMN.print_info(f"New gridprops: {new_gprops}")

        return reused_gprops, new_gprops

    def _reuse_wells(self, wells, welltype):
        """
        Identify which wells are available for reusing, reusable
        wells and new wells are returned in separate lists
        """

        new_wells = []
        reused_wells = []
        for well in wells:
            if well not in self._xtgdata[welltype]:
                new_wells.append(well)

        reused_wells = [
            value for key, value in self._xtgdata[welltype].items() if key in wells
        ]

        CMN.print_info(f"Reused wells: {[x.name for x in reused_wells]}")
        CMN.print_info(f"New wells: {new_wells}")

        return reused_wells, new_wells



"""Private module in qcforward, used to compare blocked wells with grid props.

The idea here is to compare blocked wells (cells) with actual values in order
warn user and/or stop runs if discrepancies are too large.

The resulting report will look like this:

::

             WELL COMPARE(BW:MOD) WARNRULE   STOPRULE  MATCH%  STATUS  NAMETAG
    INDEX
        0    A_6  PHIT:PHIT       any<90%    any<70%    99%     OK      MYDATA
        1    A_6  Facies:FACIES   any<90%    any<70%    91%     OK      MYDATA
        2    A_5  PHIT:PHIT       any<90%    any<70%    93%     OK      MYDATA
        3    A_5  Facies:FACIES   any<90%    any<70%    77%     WARN    MYDATA
        4    A_4  PHIT:PHIT       any<90%    any<70%    88%     WARN    MYDATA
        5    A_4  Facies:FACIES   any<90%    any<70%    73%     WARN    MYDATA
        6    all  PHIT:PHIT       all<95%    all<80%    88%     OK      MYDATA
        7    all  Facies:FACIES   all<95%    all<80%    88%     OK      MYDATA

The input spesification is on the following form if outside RMS:

::

    DATA1 = {
        "nametag": "MYDATA1",
        "verbosity": "debug",
        "path": PATH,
        "bwells": WELLFILES,
        "grid": GRIDFILE,
        "gridprops": [["PHIT", ROFFFILE], ["FACIES", ROFFFILE]],
        "compare": {"Facies": "FACIES", "PHIT": "PHIT"},  # bwname: modelname
        "actions": [
            {"warn": "anywell < 80%", "stop": "anywell < 70%"},
            {"warn": "allwells < 90%", "stop": "allwells < 80%"},
        ],
        "report": {"file": REPORT, "mode": "write"},
        "dump_yaml": SOMEYAML,
        "tolerance": 0.01
    }

If one need to have different actions for different comparisons, use e.g.:

::

    DATA2 = deepcopy(DATA1)
    DATA2["compare"] = {"VSH": "Vshale"}
    DATA2["actions"] =
        [
            {"warn": "anywell < 86%", "stop": "anywell < 75%"},
            {"warn": "allwells < 97%", "stop": "allwells < 89%"},
        ],

and rerrun.

Note that ``any`` and ``all`` will work as shortform to ``anywell`` and ``allwell``.

"""
import json
from collections import OrderedDict
from pathlib import Path
from typing import Any, Optional, Union

import pandas as pd
from jsonschema import validate

import fmu.tools
from fmu.tools._common import _QCCommon
from fmu.tools.qcdata import QCData
from fmu.tools.qcforward._qcforward import ActionsParser, QCForward, actions_validator

QCC = _QCCommon()


class _LocalData:
    def __init__(self):
        """Defining and hold data local (or special) for this routine."""
        self.actions = None
        self.compare = None
        self.infotext = "BW vs GRIDPROPS"
        self.nametag = None
        self.reportfile = None
        self.tolerance = 0.01  # tolerance when comparing
        self.show_data = None
        self.tvd_range = None

        QCC.print_debug("Initialized local data _LocalData")

    def parse_data(self, data):
        """Parsing the actual data"""

        self.nametag = data.get("nametag", "unset_nametag")
        if "report" in data:
            self.reportfile = (
                data["report"].get("file")
                if isinstance(data["report"], dict)
                else data["report"]
            )

        self.actions = actions_validator(data["actions"])

        self.compare = data["compare"]
        self.tolerance = data.get("tolerance", self.tolerance)
        self.show_data = data.get("show_data", self.show_data)
        self.tvd_range = data.get("tvd_range", self.tvd_range)
        QCC.print_debug("Parsing data is done")


class BlockedWellsVsGridProperties(QCForward):
    def run(
        self,
        data: Union[dict, str],
        reuse: Optional[bool] = False,
        project: Optional[Any] = None,
    ):
        """Main routine for evaluating blockedwells vs gridproperties

        The routine depends on existing XTGeo functions for this purpose.

        Args:
            data (dict or str): The input data either as a Python dictionary or
                a path to a YAML file
            reuse (bool or list): Reusing some "timeconsuming to read" data in the
                instance. If True, then grid and gridprops will be reused as default.
                Alternatively it can be a list for more fine grained control, e.g.
                ["grid", "gridprops", "bwells"]
            project (Union[object, str]): For usage inside RMS, None if running files

        """
        self._data = self.handle_data(data, project)
        self._validate_input(self._data, project)

        QCC.verbosity = self._data.get("verbosity", 0)

        # parse data that are special for this check
        QCC.print_info("Parsing additional data...")
        self.ldata = _LocalData()
        self.ldata.parse_data(self._data)

        # now need to retrieve blocked properties and grid properties from the "compare"
        # dictionary:
        wsettings = {"lognames": list(self.ldata.compare.keys())}

        if project:
            # inside RMS, get gridprops implicitly from compare values
            self._data["gridprops"] = list(self.ldata.compare.values())

        if not isinstance(self.gdata, QCData):
            self.gdata = QCData()

        self.gdata.parse(
            data=self._data, reuse=reuse, project=project, wells_settings=wsettings
        )

        dfr, comb = self.compare_bw_props()

        QCC.print_debug(f"Results: \n{dfr}")
        status = self.evaluate_qcreport(
            dfr, "blocked wells vs grid props", stopaction=False
        )

        # make it possible to print the underlying dataframe, either some wells (.e.g
        # the failing) or all wells. If 'fail' it will only show those lines that
        # contains FAIL
        show = self.ldata.show_data
        if show is None or show is False:
            pass
        elif isinstance(show, dict):
            if "lines" not in show or "wellstatus" not in show:
                raise ValueError(
                    f"The 'showdata' entry is in an invalid form or format: {show}"
                )

            lines = show["lines"].upper()
            wstatus = show["wellstatus"].upper()

            print(
                f"\n** Key 'show_data' is active, here showing lines with {lines} "
                f"for wells classified as {wstatus} **"
            )
            # filter out all line with word FAIL or WARN or ... , h/t HAVB
            fcomb = comb[comb.astype(str).agg("".join, axis=1).str.contains(lines)]
            if len(fcomb) > 0:
                mask = dfr["STATUS"] == wstatus
                wells = [well for well in dfr[mask]["WELL"].unique() if well != "all"]
                if wells:
                    print(f"Wells within {wstatus} criteria are: {wells}:\n")
                    print(fcomb[fcomb["WELLNAME"].isin(wells)].to_string())
                else:
                    print(f"No wells within {wstatus} criteria")
            else:
                print(f"No lines are matching {lines}. Wrong input?:\n")

        else:

            print("Show all well cells for all wells:")
            if len(comb) > 0:
                print(comb.to_string())

        if status == "STOP":
            QCC.force_stop("STOP criteria is found!")

    def compare_bw_props(self) -> pd.DataFrame:
        """Given data, do a comparison of blcked wells cells vs props, via XTGeo."""

        # dataframe for the blocked wells
        dfbw = self.gdata.bwells.get_dataframe()
        if self._gdata.project is not None:
            # when parsing blocked wells from RMS, cell indices starts from 0, not  1
            dfbw["I_INDEX"] += 1
            dfbw["J_INDEX"] += 1
            dfbw["K_INDEX"] += 1

        # filtering on depth tvd_range:
        if self.ldata.tvd_range and isinstance(self.ldata.tvd_range, list):
            zmin = self.ldata.tvd_range[0]
            zmax = self.ldata.tvd_range[1]
            if zmin >= zmax:
                raise ValueError("The zmin value >= zmax in 'tvd_range'")
            dfbw = dfbw[dfbw["Z_TVDSS"] >= zmin]
            dfbw = dfbw[dfbw["Z_TVDSS"] <= zmax]
            if dfbw.empty:
                raise RuntimeError(
                    f"No wells left after tvd_range: {self.ldata.tvd_range}"
                )

        # dataframe for the properties, need some processing (column names)
        dfprops = self.gdata.gridprops.get_dataframe(ijk=True, grid=self.gdata.grid)
        dfprops = dfprops.rename(
            columns={"IX": "I_INDEX", "JY": "J_INDEX", "KZ": "K_INDEX"}
        )

        # merge the dataframe on I J K index
        comb = pd.merge(
            dfbw,
            dfprops,
            how="inner",
            on=["I_INDEX", "J_INDEX", "K_INDEX"],
            suffixes=("__bw", "__model"),  # in case the names are equal -> add suffix
        )
        QCC.print_debug("Made a combined dataframe!")
        QCC.print_debug(f"\n {comb}")

        diffs = {}

        # compare the relevant properties
        for bwprop, modelprop in self._ldata.compare.items():
            usebwprop = bwprop if bwprop != modelprop else bwprop + "__bw"
            usemodelprop = modelprop if bwprop != modelprop else modelprop + "__model"
            dname = bwprop + ":" + modelprop
            dnameflag = dname + "_flag"
            comb = self._eval_tolerance(comb, usebwprop, usemodelprop, dname, dnameflag)
            diffs[dname] = dnameflag

        return self._evaluate_diffs(comb, diffs), comb

    def _eval_tolerance(self, df_in, bwprop, modelprop, diffname, diffnameflag):
        """Make a flag log for diffs based on tolerance input."""
        comb = df_in.copy()
        tol = self.ldata.tolerance

        relative = isinstance(tol, dict) and "rel" in tol
        tolerance = tol if isinstance(tol, float) else list(tol.values())[0]

        comb[diffname] = comb[bwprop] - comb[modelprop]
        comb[diffnameflag] = "MATCH"
        if relative:  # adjust relative to be weighted on mean() value
            comb[bwprop + "_mean"] = comb[bwprop].mean()
            comb[diffname + "_rel"] = comb[diffname] / comb[bwprop + "_mean"]
            comb.loc[abs(comb[diffname + "_rel"]) > tolerance, diffnameflag] = "FAIL"
        else:
            comb.loc[abs(comb[diffname]) > tolerance, diffnameflag] = "FAIL"

        return comb

    def _evaluate_diffs(self, comb, diffs) -> pd.DataFrame:

        result: OrderedDict = OrderedDict(
            [
                ("WELL", []),
                ("COMPARE(BW:MODEL)", []),
                ("WARNRULE", []),
                ("STOPRULE", []),
                ("MATCH%", []),
                ("STATUS", []),
            ]
        )

        wells = list(comb["WELLNAME"].unique())
        wells.append("all")

        QCC.print_info("Compare per well...")
        for wname in wells:
            subset = comb[comb["WELLNAME"] == wname]
            for diff, flag in diffs.items():
                result["WELL"].append(wname)
                result["COMPARE(BW:MODEL)"].append(diff)
                if wname != "all":
                    match = subset[flag].value_counts(normalize=True)["MATCH"] * 100.0
                else:
                    match = comb[flag].value_counts(normalize=True)["MATCH"] * 100.0

                result["MATCH%"].append(match)
                status = "OK"

                for therule in self.ldata.actions:

                    warnrule = ActionsParser(
                        therule.get("warn", None), mode="warn", verbosity=QCC.verbosity
                    )
                    stoprule = ActionsParser(
                        therule.get("stop", None), mode="stop", verbosity=QCC.verbosity
                    )

                    for _, issue in enumerate([warnrule, stoprule]):

                        if wname != "all" and not issue.all:
                            rulename = issue.mode.upper() + "RULE"
                            result[rulename].append(issue.expression)
                            if issue.compare == "<" and match < issue.limit:
                                status = issue.mode.upper()

                        elif wname == "all" and issue.all:
                            rulename = issue.mode.upper() + "RULE"
                            result[rulename].append(issue.expression)

                            if issue.compare == "<" and match < issue.limit:
                                status = issue.mode.upper()
                result["STATUS"].append(status)

        dfr = self.make_report(
            result, reportfile=self.ldata.reportfile, nametag=self.ldata.nametag
        )
        QCC.print_info("Dataframe is created")
        return dfr

    @staticmethod
    def _validate_input(data, project):
        """Validate data against JSON schemas, TODO complete schemas"""

        spath = Path(fmu.tools.__file__).parent / "qcforward" / "_schemas"

        schemafile = "bw_vs_gridprops_asfile.json"

        if project:
            schemafile = "bw_vs_gridprops_asroxapi.json"

        with open((spath / schemafile), "r", encoding="utf8") as thisschema:
            schema = json.load(thisschema)

        validate(instance=data, schema=schema)



"""Script for generating a design matrix from config input"""

import argparse
from pathlib import Path

from fmu.tools.sensitivities import DesignMatrix, excel2dict_design


def get_parser():

    parser = argparse.ArgumentParser(
        description="Generate design matrix to be used with ert DESIGN2PARAMS",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "config", type=str, help="Input design config filename in Excel format"
    )
    parser.add_argument(
        "destination",
        type=str,
        nargs="?",
        help="Destination filename for design matrix",
        default="generateddesignmatrix.xlsx",
    )
    parser.add_argument(
        "--designinput",
        type=str,
        help="Alternative sheetname for the worksheet designinput",
        default="designinput",
    )
    parser.add_argument(
        "--defaultvalues",
        type=str,
        help="Alternative sheetname for worksheet defaultvalues",
        default="defaultvalues",
    )
    parser.add_argument(
        "--general_input",
        type=str,
        help="Alternative sheetname for the worksheet general_input",
        default="general_input",
    )

    return parser


def main():
    """fmudesign is a command line utility for generating design matrices

    Wrapper for the the fmu.tools.sensitivities module"""

    parser = get_parser()
    args = parser.parse_args()

    # Defaulted options should be reset to None, so that the other
    # defaulting level inside _excel2dict can do its work.
    if args.designinput == parser.get_default("designinput"):
        args.designinput = None
    if args.defaultvalues == parser.get_default("defaultvalues"):
        args.defaultvalues = None
    if args.general_input == parser.get_default("general_input"):
        args.general_input = None

    sheetnames = dict()
    if args.designinput:
        sheetnames["designinput"] = args.designinput
    if args.defaultvalues:
        sheetnames["defaultvalues"] = args.defaultvalues
    if args.general_input:
        sheetnames["general_input"] = args.general_input

    if sheetnames:
        print("Worksheets changed from default:")
        print(sheetnames)

    if isinstance(args.config, str):
        if not Path(args.config).is_file():
            raise IOError("Input file {} does not exist".format(args.config))
        input_dict = excel2dict_design(args.config, sheetnames)

    if args.config == args.destination:
        raise IOError(
            'Identical name "{}" have been provided for the input'
            "file and the output file. "
            " Exiting.....".format(args.config)
        )

    design = DesignMatrix()

    design.generate(input_dict)

    Path(args.destination).parent.mkdir(exist_ok=True, parents=True)

    design.to_xlsx(args.destination)


if __name__ == "__main__":
    main()



import pytest

from fmu.tools.fipmapper import fipmapper


@pytest.mark.parametrize(
    "input_dict, expected_inverse",
    [
        ({}, {}),
        ({"foo": "bar"}, {"bar": ["foo"]}),
        ({"foo": "1", "bar": "2"}, {"1": ["foo"], "2": ["bar"]}),
        ({"foo": "1", "bar": "1"}, {"1": ["bar", "foo"]}),
        ({"foo": 1, "bar": 1}, {1: ["bar", "foo"]}),
        ({1: "foo", 2: "foo"}, {"foo": [1, 2]}),
        ({2: "foo", 1: "foo"}, {"foo": [1, 2]}),
        (
            {"foo": [1, 2], "bar": [3, 4]},
            {1: ["foo"], 2: ["foo"], 3: ["bar"], 4: ["bar"]},
        ),
        (
            {"foo": [1, 2], "bar": [3, 4], "Totals": [1, 2, 3, 4]},
            {
                1: ["Totals", "foo"],
                2: ["Totals", "foo"],
                3: ["Totals", "bar"],
                4: ["Totals", "bar"],
            },
        ),
    ],
)
def test_invert_map(input_dict, expected_inverse):
    assert fipmapper.invert_map(input_dict) == expected_inverse


def test_invert_map_skipstring():
    input_dict = {"foo": [1, 2], "bar": [3, 4], "Totals": [1, 2, 3, 4]}
    assert fipmapper.invert_map(input_dict, skipstring="Totals") == {
        1: ["foo"],
        2: ["foo"],
        3: ["bar"],
        4: ["bar"],
    }


def test_fipmapper_empty():
    mapper = fipmapper.FipMapper()
    assert mapper.has_region2fip is False
    assert mapper.has_zone2fip is False
    assert mapper.has_fip2region is False
    assert mapper.has_fip2zone is False


def test_fipmapper():
    mapper = fipmapper.FipMapper(
        mapdata={"fipnum2region": {1: "West-Brent", 2: "East-Sognefjord"}}
    )
    assert mapper.fip2region(1) == ["West-Brent"]
    assert mapper.fip2region(2) == ["East-Sognefjord"]
    assert mapper._fips2regions([1, 2]) == [["West-Brent"], ["East-Sognefjord"]]
    assert mapper.region2fip("West-Brent") == [1]
    assert mapper._regions2fips(["West-Brent"]) == [[1]]
    assert mapper._regions2fips(["West-Brent", "East-Sognefjord"]) == [[1], [2]]


def test_fipmapper_zones():
    mapper = fipmapper.FipMapper(
        mapdata={"fipnum2zone": {1: "Upper", 2: "Middle", 3: "Middle"}}
    )
    assert mapper.fip2zone(1) == ["Upper"]
    assert mapper.fip2zone(2) == ["Middle"]
    assert mapper.fip2zone(3) == ["Middle"]
    assert mapper.zone2fip("Upper") == [1]
    assert mapper.zone2fip("Middle") == [2, 3]


def test_integer_regions():
    """Regions are sometimes integer, and then they will
    typically be returned as integers from the yaml parsing"""
    mapper = fipmapper.FipMapper(mapdata={"fipnum2region": {1: 1, 2: 2}})
    assert mapper.fip2region(1) == [1]
    assert mapper.fip2region(2) == [2]
    assert mapper.region2fip(1) == [1]
    assert mapper.region2fip(2) == [2]


def test_integer_zones():
    """Should also allow using integers for zones. Maybe
    the integer is actually the k index"""
    mapper = fipmapper.FipMapper(mapdata={"fipnum2zone": {1: 1, 2: 2}})
    assert mapper.fip2zone(1) == [1]
    assert mapper.fip2zone(2) == [2]
    assert mapper.zone2fip(1) == [1]
    assert mapper.zone2fip(2) == [2]


def test_mixed_datatypes():
    """Mixed ints/strs in regions and zones"""
    mapper = fipmapper.FipMapper(
        mapdata={"fipnum2region": {1: 1, 2: "B"}, "fipnum2zone": {1: 1, 2: "L"}}
    )
    assert mapper.fip2region(1) == [1]
    assert mapper.fip2region(2) == ["B"]
    assert mapper.region2fip(1) == [1]
    assert mapper.region2fip("B") == [2]

    assert mapper.fip2zone(1) == [1]
    assert mapper.fip2zone(2) == ["L"]
    assert mapper.zone2fip(1) == [1]
    assert mapper.zone2fip("L") == [2]


def test_fipmapper_regzone2fip():
    mapper = fipmapper.FipMapper(
        mapdata={
            "fipnum2zone": {1: "Upper", 2: "Middle", 3: "Upper", 4: "Middle"},
            "fipnum2region": {
                1: "West-Brent",
                2: "West-Brent",
                3: "East-Sognefjord",
                4: "East-Sognefjord",
            },
        }
    )
    assert mapper.regzone2fip("West-Brent", "Upper") == [1]
    assert mapper.regzone2fip("West-Brent", "Middle") == [2]
    assert mapper.regzone2fip("East-Sognefjord", "Upper") == [3]
    assert mapper.regzone2fip("East-Sognefjord", "Middle") == [4]


@pytest.mark.parametrize(
    "map_data, expected_regions, expected_zones, expected_fipnums",
    [
        ({"region2fipnum": {"A": 1}, "zone2fipnum": {"U": 1}}, ["A"], ["U"], [1]),
        (
            {"region2fipnum": {"A": [1, 2]}, "zone2fipnum": {"U": 1}},
            ["A"],
            ["U"],
            [1, 2],
        ),
        (
            # Test sorting
            {"region2fipnum": {"A": [2, 1]}, "zone2fipnum": {"U": [2, 1]}},
            ["A"],
            ["U"],
            [1, 2],
        ),
        (
            # Test sorting with integer regions
            {"region2fipnum": {10: [3], 1: [2, 1]}, "zone2fipnum": {"U": [2, 1, 3]}},
            [1, 10],
            ["U"],
            [1, 2, 3],
        ),
        (
            {
                "region2fipnum": {"A": [2, 1], "B": 3},
                "zone2fipnum": {"U": [1], "L": [3, 2]},
            },
            ["A", "B"],
            ["L", "U"],
            [1, 2, 3],
        ),
        pytest.param({}, [], [], [], marks=pytest.mark.xfail(raises=AssertionError)),
        pytest.param(
            {"region2fipnum": {"A": 1}},
            ["A"],
            [],
            [1],
            # get_zones fails here.
            marks=pytest.mark.xfail(raises=AssertionError),
        ),
        pytest.param(
            {"zone2fipnum": {"U": 1}},
            [],
            ["U"],
            [1],
            # get_regions fails here.
            marks=pytest.mark.xfail(raises=AssertionError),
        ),
        pytest.param(
            {"zone2fipnum": {"U": "fip1"}, "region2fipnum": {"W": "fip1"}},
            ["W"],
            ["U"],
            ["fip1"],
            marks=pytest.mark.xfail(raises=TypeError),
        ),
    ],
)
def test_get_regions_zones_fipnums(
    map_data, expected_regions, expected_zones, expected_fipnums
):
    """Test the three functions get_regions, get_zones and get_fipnums"""
    mapper = fipmapper.FipMapper(mapdata=map_data)
    assert mapper.get_regions() == expected_regions
    assert mapper.get_zones() == expected_zones
    assert mapper.get_fipnums() == expected_fipnums


@pytest.mark.parametrize(
    "input_dict, expected_dict",
    [
        ({}, {}),
        ({"FIPNUM": {}}, {}),
        ({"FIPNUM": 0}, {}),
        ({"FIPNUM": {"groups": {}}}, {}),
        ({"FOO": 0}, {}),
        # First test with the webviz format @march 2021, with "groups"
        # as a required level:
        (
            {"FIPNUM": {"groups": {"REGION": {"west": 1}}}},
            {"region2fipnum": {"west": 1}},
        ),
        (
            {"FIPNUM": {"groups": {"REGION": {"west": [1]}, "ZONE": {"lower": [2]}}}},
            {"region2fipnum": {"west": [1]}, "zone2fipnum": {"lower": [2]}},
        ),
        # The "groups" level might disappear in webviz format, so ensure
        # we also support that when/if it happens:
        (
            {"FIPNUM": {"REGION": {"west": 1}}},
            {"region2fipnum": {"west": 1}},
        ),
        (
            {"FIPNUM": {"REGION": {"west": [1]}, "ZONE": {"lower": [2]}}},
            {"region2fipnum": {"west": [1]}, "zone2fipnum": {"lower": [2]}},
        ),
    ],
)
def test_webviz_to_prtvol2csv(input_dict, expected_dict):
    print(fipmapper.webviz_to_prtvol2csv(input_dict))
    assert fipmapper.webviz_to_prtvol2csv(input_dict) == expected_dict



""" Private class in qcproperties """

from fmu.tools._common import _QCCommon

QCC = _QCCommon()


class ConfigParser:
    """
    Class for parsing and preparing the input data for extracting statistics
    with QCProperties. The input data is formatted and grouped into relevant
    attributes based on where it will be utilized.
    """

    def __init__(
        self,
        data: dict,
    ):

        QCC.verbosity = data.get("verbosity", 0)

        if "csvfile" in data:
            raise KeyError(
                "Use of 'csvfile' keyword in data is deprecated. "
                "To output a csv-file use the to_csv method on "
                "the QCProperties() instance instead!"
            )

        self._aggregation_controls: dict = dict(
            properties=[],
            selectors=[],
            weights={},
        )

        self._prop2df_controls: dict = dict(
            unique_parameters=[],
            properties_input_names=[],
            selectors_input_names=[],
            filters={},
            name_mapping={},
            usercodes={},
        )

        self._data_loading_input: dict = dict(pfiles={}, pdates={})

        # set data loading input
        for item in ["grid", "wells", "bwells", "path", "verbosity"]:
            if item in data:
                self._data_loading_input[item] = data[item]

        # set aggregation controls and parameter data
        self._parse_properties(data["properties"])
        self._parse_selectors(data.get("selectors", {}))
        self._parse_filters(data.get("filters", {}))
        self._aggregation_controls["selector_combos"] = data.get(
            "selector_combos", True
        )
        self._aggregation_controls["output_percentage"] = data.get(
            "output_percentage", False
        )
        self._aggregation_controls["verbosity"] = QCC.verbosity

    # ==================================================================================
    # Properties
    # ==================================================================================

    @property
    def data_loading_input(self) -> dict:
        """Attribute to use for loading data to XTGeo"""
        return self._data_loading_input

    @property
    def aggregation_controls(self) -> dict:
        """Attribute to use for statisticts aggregation"""
        return self._aggregation_controls

    @property
    def prop2df_controls(self) -> dict:
        """Attribute to use for creating dataframe from properties"""
        return self._prop2df_controls

    # ==================================================================================
    # Hidden class methods
    # ==================================================================================

    def _parse_properties(self, properties):
        """Add property data to relevant attributes"""

        if isinstance(properties, list):
            properties = {param: {"name": param} for param in properties}

        for column_name, values in properties.items():
            name = values["name"]
            self._aggregation_controls["properties"].append(column_name)
            self._prop2df_controls["properties_input_names"].append(name)
            self._prop2df_controls["name_mapping"][name] = column_name
            self._add_to_parameters(name)

            if "weight" in values:
                self._aggregation_controls["weights"][column_name] = values["weight"]
                self._add_to_parameters(values["weight"])
            if "range" in values:
                self._prop2df_controls["filters"][name] = {"range": values["range"]}
            if "pfile" in values:
                self._data_loading_input["pfiles"][name] = values["pfile"]

        QCC.print_debug(f"properties: {properties}")

    def _parse_selectors(self, selectors):
        """Add selector data to relevant attributes"""

        if isinstance(selectors, list):
            selectors = {param: {"name": param} for param in selectors}

        for column_name, values in selectors.items():
            name = values["name"]
            self._aggregation_controls["selectors"].append(column_name)
            self._prop2df_controls["selectors_input_names"].append(name)
            self._prop2df_controls["name_mapping"][name] = column_name
            self._add_to_parameters(name)

            if "include" in values and "exclude" in values:
                raise ValueError("can't both include and exclude values in filtering")
            if "include" in values:
                self._prop2df_controls["filters"][name] = {"include": values["include"]}
            if "exclude" in values:
                self._prop2df_controls["filters"][name] = {"exclude": values["exclude"]}
            if "codes" in values:
                self._prop2df_controls["usercodes"][name] = values["codes"]
            if "pfile" in values:
                self._data_loading_input["pfiles"][name] = values["pfile"]

        QCC.print_debug(f"selectors: {selectors}")

    def _parse_filters(self, filters):
        """Add additional filters to relevant attributes"""
        for name, values in filters.items():
            self._add_to_parameters(name)

            if "pfile" in values:
                self._data_loading_input["pfiles"][name] = values["pfile"]

            # support using a selector prop as filter. If the selctor
            # has filters specified in its values, they will be ignored
            if name in self._prop2df_controls["filters"]:
                QCC.give_warn(
                    f"Filters for {name} found both in 'filters' and 'selectors' "
                    "or 'properties'. The filter on the selector/property is ignored."
                )
            self._prop2df_controls["filters"][name] = values

        # Filter format check
        for values in self._prop2df_controls["filters"].values():
            if "include" in values:
                if isinstance(values["include"], str):
                    values["include"] = [values["include"]]
                if not all(isinstance(item, str) for item in values["include"]):
                    values["include"] = [str(item) for item in values["include"]]

            if "exclude" in values:
                if isinstance(values["exclude"], str):
                    values["exclude"] = [values["exclude"]]
                if not all(isinstance(item, str) for item in values["exclude"]):
                    values["exclude"] = [str(item) for item in values["exclude"]]

            if "range" in values:
                if not (
                    isinstance(values["range"], list) and len(values["range"]) == 2
                ):
                    raise TypeError(
                        "Filter range must be input as list with two values"
                    )

        QCC.print_debug(f"Filters: {self._prop2df_controls['filters']}")

    def _add_to_parameters(self, param):
        """Add parameter to list of unique parameters"""
        if param not in self._prop2df_controls["unique_parameters"]:
            self._prop2df_controls["unique_parameters"].append(param)



"""
This private module in qcforward is used to check grid quality
"""

from pathlib import Path
import json
from copy import deepcopy
from collections import OrderedDict

from jsonschema import validate
import fmu.tools


from fmu.tools._common import _QCCommon
from fmu.tools.qcdata import QCData
from fmu.tools.qcforward._qcforward import QCForward, ActionsParser


QCC = _QCCommon()

UNDEF = float("nan")


class _LocalData:
    def __init__(self):
        """Defining and hold data local for this routine"""

        self.actions = None
        self.infotext = "GRID QUALITY"
        self.nametag = None
        self.reportfile = None
        self.writeicon = False

    def parse_data(self, data):
        """Parsing the actual data"""

        # TODO: verify and qc
        self.nametag = data.get("nametag", "unset_nametag")
        if "report" in data:
            self.reportfile = (
                data["report"].get("file")
                if isinstance(data["report"], dict)
                else data["report"]
            )

        self.actions = data["actions"]
        self.writeicon = data.get("writeicon", False)


class GridQuality(QCForward):
    def run(self, data, reuse=False, project=None):
        """Main routine for evaluating grid quality and stop/warn if too bad

        The routine depends on existing XTGeo functions for this purpose.

        Args:
            data (dict or str): The input data either as a Python dictionary or
                a path to a YAML file
            reuse (bool or list): Reusing some "timeconsuming to read" data in the
                instance. If True, then grid and gridprops will be reused as default.
                Alternatively it can be a list for more fine grained control, e.g.
                ["grid", "gridprops", "wells"]
            project (Union[object, str]): For usage inside RMS, None if running files

        """
        self._data = self.handle_data(data, project)
        self._validate_input(self._data, project)

        QCC.verbosity = self._data.get("verbosity", 0)

        # parse data that are special for this check
        QCC.print_info("Parsing additional data...")
        self.ldata = _LocalData()
        self.ldata.parse_data(data)

        if isinstance(self.gdata, QCData):
            self.gdata.parse(data=data, reuse=reuse, project=project)
        else:
            self.gdata = QCData()
            self.gdata.parse(data)

        dfr = self.check_gridquality()
        QCC.print_debug(f"Results: \n{dfr}")

        self.evaluate_qcreport(dfr, "grid quality")

    def check_gridquality(self):
        """
        Given data, do check of gridquality via XTGeo

        Final result will be a table like this::

                                WARNRULE       WARN%  STOPRULE      STOP%  STATUS...
          GRIDQUALITY
          minangle_top_base[0]  all>10%ifx<60  13.44  all>0%ifx<40  2.32   WARN
          collapsed[0]          all>15%        12.25  allcells>30%  0.0    OK
        """

        # get properties via XTGeo method get_gridquality_properties()
        gqc = self.gdata.grid.get_gridquality_properties()

        actions = self.ldata.actions
        if actions is None:
            raise ValueError("No actions are defined for grid quality")

        result = OrderedDict(
            [
                ("GRIDQUALITY", []),
                ("WARNRULE", []),
                ("WARN%", []),
                ("STOPRULE", []),
                ("STOP%", []),
                ("STATUS", []),
            ]
        )

        for prop in gqc.props:
            # gqc.props is a list of all gridquality properties, but not all of these
            # are defined in input actions.

            therules = actions.get(prop.name, None)

            if self.data["project"] and self.ldata.writeicon and therules is not None:
                QCC.print_info(f"Write icon in RMS for {prop.name}")
                prop.to_roxar(self.data["project"], self.data["grid"], prop.name)

            if therules is None:
                continue

            for numrule, therule in enumerate(therules):

                warnrule = ActionsParser(
                    therule.get("warn", None), mode="warn", verbosity=QCC.verbosity
                )
                stoprule = ActionsParser(
                    therule.get("stop", None), mode="stop", verbosity=QCC.verbosity
                )

                QCC.print_debug(f"WARN RULE {warnrule.status}")
                QCC.print_debug(f"STOP RULE {stoprule.status}")

                # if stoprule is None or warnrule is None:
                #     raise ValueError("Rules for both warn and stop must be defined")

                result["GRIDQUALITY"].append(f"{prop.name}[{numrule}]")

                status = "OK"
                for issue in [warnrule, stoprule]:
                    status, result = self._evaluate_allcells(
                        issue, result, prop, status
                    )

                result["STATUS"].append(status)

        dfr = self.make_report(
            result, reportfile=self.ldata.reportfile, nametag=self.ldata.nametag
        )
        dfr.set_index("GRIDQUALITY", inplace=True)
        return dfr

    @staticmethod
    def _evaluate_allcells(issue, inresult, prop, instatus):
        """Evaluation of all cells per issue (warn or stop) given the criteria."""

        result = deepcopy(inresult)

        if issue.status is None:
            result[issue.mode.upper() + "%"].append(UNDEF)
            result[issue.mode.upper() + "RULE"].append(UNDEF)
            status = "OK"
            return status, result

        ncell = prop.values.count()

        if issue.given == "<":
            nbyrule = (prop.values < issue.criteria).sum()
        elif issue.given == ">":
            nbyrule = (prop.values > issue.criteria).sum()
        else:
            # e.g. discrete qual parameters such as 'faulted' have only 0 or 1 values
            nbyrule = (prop.values > 0).sum()

        actualpercent = 100.0 * nbyrule / ncell

        result[issue.mode.upper() + "%"].append(actualpercent)
        result[issue.mode.upper() + "RULE"].append(issue.expression)

        if issue.compare == ">" and actualpercent > issue.limit:
            status = issue.mode.upper()
        elif issue.compare == "<" and actualpercent < issue.limit:
            status = issue.mode.upper()
        else:
            status = instatus

        return status, result

    @staticmethod
    def _validate_input(data, project):
        """Validate data against JSON schemas, TODO complete schemas"""

        spath = Path(fmu.tools.__file__).parent / "qcforward" / "_schemas"

        schemafile = "gridquality_asfile.json"

        if project:
            schemafile = "gridquality_asroxapi.json"

        with open((spath / schemafile), "r") as thisschema:
            schema = json.load(thisschema)

        validate(instance=data, schema=schema)



import pathlib
from os.path import abspath
import pytest
import pandas as pd

from fmu.tools import qcforward as qcf

SOMEYAML = abspath("/tmp/somefile.yml")
REPORT = abspath("/tmp/somefile.csv")


@pytest.fixture(name="data")
def fixture_data():
    return {
        "path": abspath("../xtgeo-testdata/3dgrids/reek/"),
        "grid": "reek_sim_grid.roff",
        "report": REPORT,
        "verbosity": 1,
    }


def test_simple_action(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "warn_outside": [0.18, 0.25],
            "stop_outside": [0, 1],
            "description": "test",
        },
    ]

    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    dfr = pd.read_csv(REPORT)

    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["STATUS"] == "WARN"
    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["VALUE"] == pytest.approx(
        0.1677, 0.001
    )

    pathlib.Path(REPORT).unlink()


def test_action_with_disc_and_cont_props(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "warn_outside": [0.18, 0.25],
            "stop_outside": [0, 1],
            "description": "test",
        },
        {
            "property": "reek_sim_facies2.roff",
            "codename": "SHALE",
            "warn_outside": [0.3, 0.7],
            "stop_outside": [0, 1],
            "description": "test2",
        },
        {
            "property": "reek_sim_facies2.roff",
            "codename": "SHALE",
            "selectors": {"reek_sim_zone.roff": "Below_Top_reek"},
            "warn_outside": [0.3, 0.7],
            "stop_outside": [0, 1],
            "description": "test3",
        },
    ]

    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    dfr = pd.read_csv(REPORT)

    assert dfr.loc[dfr["PROPERTY"] == "reek_sim_poro.roff"].iloc[0]["STATUS"] == "WARN"
    assert dfr.loc[dfr["PROPERTY"] == "reek_sim_poro.roff"].iloc[0][
        "VALUE"
    ] == pytest.approx(0.1677, 0.001)
    assert dfr.loc[
        (dfr["PROPERTY"] == "reek_sim_facies2.roff") & (dfr["DESCRIPTION"] == "test2")
    ].iloc[0]["VALUE"] == pytest.approx(0.585, abs=0.001)

    pathlib.Path(REPORT).unlink()


def test_multiple_actions(data):

    zones_stop = [
        ["Below_Top_reek", [0.1, 0.3]],
        ["Below_Mid_reek", [0.1, 0.25]],
        ["Below_Low_reek", [0.1, 0.20]],
    ]

    actions = []
    for zstop in zones_stop:
        actions.append(
            {
                "property": "reek_sim_poro.roff",
                "selectors": {"reek_sim_zone.roff": zstop[0]},
                "stop_outside": zstop[1],
            }
        )

    data["actions"] = actions
    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    dfr = pd.read_csv(REPORT)
    print(dfr)
    pathlib.Path(REPORT).unlink()


def test_action_with_selectors(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "selectors": {"reek_sim_zone.roff": "Below_Mid_reek"},
            "warn_outside": [0.17, 0.25],
            "stop_outside": [0, 1],
        },
    ]

    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    dfr = pd.read_csv(REPORT)

    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["STATUS"] == "WARN"
    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["VALUE"] == pytest.approx(
        0.1606, 0.001
    )

    pathlib.Path(REPORT).unlink()


def test_action_with_filters(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "filters": {
                "reek_sim_zone.roff": {"exclude": ["Below_Top_reek", "Below_Low_reek"]},
                "reek_sim_facies2.roff": {
                    "include": ["FINESAND", "COARSESAND"],
                },
            },
            "warn_outside": [0.17, 0.25],
            "stop_outside": [0, 1],
        },
    ]

    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    dfr = pd.read_csv(REPORT)

    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["STATUS"] == "OK"
    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["VALUE"] == pytest.approx(
        0.2384, 0.001
    )

    pathlib.Path(REPORT).unlink()


def test_action_with_filters_and_selectors(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "selectors": {"reek_sim_zone.roff": "Below_Mid_reek"},
            "filters": {
                "reek_sim_facies2.roff": {
                    "include": ["FINESAND", "COARSESAND"],
                },
            },
            "warn_outside": [0.17, 0.25],
            "stop_outside": [0, 1],
        },
    ]

    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    dfr = pd.read_csv(REPORT)

    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["STATUS"] == "OK"
    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["VALUE"] == pytest.approx(
        0.2384, 0.001
    )

    pathlib.Path(REPORT).unlink()


def test_actions_shall_stop(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "warn_outside": [0.17, 0.4],
            "stop_outside": [0.20, 1],
        }
    ]

    qcjob = qcf.GridStatistics()
    with pytest.raises(SystemExit):
        qcjob.run(data)


def test_actions_shall_stop_no_warnlimits(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "stop_outside": [0.20, 1],
        }
    ]

    qcjob = qcf.GridStatistics()
    with pytest.raises(SystemExit):
        qcjob.run(data)


def test_actions_with_selectors(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "selectors": {
                "reek_sim_zone.roff": "Below_Top_reek",
                "reek_sim_facies2.roff": "COARSESAND",
            },
            "warn_outside": [0.17, 0.25],
            "stop_outside": [0, 1],
            "calculation": "Avg",
        },
    ]

    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    dfr = pd.read_csv(REPORT)

    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["STATUS"] == "WARN"
    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["VALUE"] == pytest.approx(
        0.3117, 0.001
    )

    pathlib.Path(REPORT).unlink()


def test_yaml_dump(data):

    data["actions"] = [
        {
            "property": "reek_sim_poro.roff",
            "warn_outside": [0.18, 0.25],
            "stop_outside": [0, 1],
        },
    ]

    data["dump_yaml"] = SOMEYAML
    qcjob = qcf.GridStatistics()
    qcjob.run(data)

    # now read the dump file:
    qcjob.run(data=SOMEYAML)

    dfr = pd.read_csv(REPORT)

    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["STATUS"] == "WARN"
    assert dfr.loc[dfr["CALCULATION"] == "Avg"].iloc[0]["VALUE"] == pytest.approx(
        0.1677, 0.001
    )

    pathlib.Path(REPORT).unlink()
    pathlib.Path(SOMEYAML).unlink()
